/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef base_h
#define base_h

// everything includes base.h, this is the minimal include

#include <ta/ta_group.h>
#include <pdp/declare.h>
#include <pdp/pdp_TA_type.h> 
#include <ta_misc/tdgeometry.h>
#include <ta_misc/win_base.h>
#include <setjmp.h>

class ColorScaleSpec;

// all of these functions are cached out in pdpshell.cc
 
class pdpMisc {
  // #NO_TOKENS miscellaneous things for pdp
public:
  static PDPRoot*	root;		// root of pdp class hierarchy
  static char*		defaults_str;
  // string representation of basic defaults compiled into executable
  static String_Array	proj_to_load;	// list of projects to load 
#ifndef __MAKETA__
  static jmp_buf 	proc_begin; 	// process jump pointer
#endif

  static void (*Init_Hook)();	 	// #READ_ONLY set this in user's main to init ta, etc.

  static int	Main(int argc, char* argv[]); 
  // #IGNORE the pdp main startup function: call this from user's main

  static int	WaitProc();
  // waiting process 

  static void 	SaveRecoverFile(int err);
  // error handling function that saves a recover file when system crashes 

  static ColorScaleSpec* GetDefaultColor();
  // gets the default color scale 
};


//////////////////////////////////////////////////
//	PDPView:  view for pdp system		//
//////////////////////////////////////////////////

class PDPView : public WinView {
  // #NO_TOKENS pdp version of a win view
public:
  void		AddNotify(TAPtr ud);
  void		RemoveNotify(TAPtr ud);
  TypeDef* 	UpdaterType()		{ return &TA_SchedProcess; }

  void	Initialize()	{ };
  void 	Destroy()	{ };
  TA_BASEFUNS(PDPView);
};

// PDPMGroup will update the project upon changes to any objects
class PDPMGroup : public MenuGroup_impl {
public:
  void		UpdateMenu();

  void 	Initialize();
  void 	Destroy();
  TA_BASEFUNS(PDPMGroup);
};

//////////////////////////////////////////////////
//	PosGroup/MGroup: groups with positions	//
//////////////////////////////////////////////////

#ifdef USE_TEMPLATE_GROUPS

#define PosGroup_of(T)							      \
class T ## _Group : public taGroup<T> {				      \
public:									      \
  TDCoord	pos;		/* Position of Group */	     		      \
  void	Initialize() 		{ };					      \
  void 	Destroy()		{ };					      \
  void	InitLinks()		{ taGroup<T>::InitLinks(); taBase::Own(pos,this); }\
  void  Copy_(const T ## _Group& cp)		{ pos = cp.pos; }		      \
  void  Copy(const T ## _Group& cp)		{ taGroup<T>::Copy(cp); Copy_(cp); } \
  TA_BASEFUNS(T ## _Group);						      \
}

#define PosMGroup_of(T)							      	\
class T ## _MGroup : public MenuGroup<T> {				      	\
public:									      	\
  TDCoord	pos;		/* Position of Group */	     			\
  void	Initialize() 		{ };					      	\
  void 	Destroy()		{ };					      	\
  void	InitLinks()		{ MenuGroup<T>::InitLinks(); taBase::Own(pos,this); }\
  void  Copy_(const T ## _MGroup& cp)	{ pos = cp.pos; }		        	\
  void  Copy(const T ## _MGroup& cp)	{ MenuGroup<T>::Copy(cp); Copy_(cp); } 		\
  TA_BASEFUNS(T ## _MGroup);						      	\
}

#else // USE_TEMPLATE_GROUPS

class PosGroup : public taBase_Group { // ##NO_TOKENS
public:
  TDCoord	pos;		// Position of Group
  
  void	Initialize() 		{ };
  void 	Destroy()		{ };
  void	InitLinks()		{ taBase_Group::InitLinks(); taBase::Own(pos,this); }
  void  Copy_(const PosGroup& cp)	{ pos = cp.pos; }
  COPY_FUNS(PosGroup, taBase_Group);
  TA_BASEFUNS(PosGroup);
};

#define PosGroup_of(T)							      \
class T ## _Group : public PosGroup {					      \
public:									      \
  void	Initialize() 		{ SetBaseType(&TA_ ## T); }		      \
  void 	Destroy()		{ };					      \
  TA_BASEFUNS(T ## _Group);						      \
}

class PosMGroup : public PDPMGroup {
public:
  TDCoord	pos;		// Position of Group

  void	Initialize() 		{ };
  void 	Destroy()		{ };
  void	InitLinks()		{ PDPMGroup::InitLinks(); taBase::Own(pos,this); }
  void  Copy_(const PosMGroup& cp)	{ pos = cp.pos; }
  void  Copy(const PosMGroup& cp)	{ PDPMGroup::Copy(cp); Copy_(cp); }
  TA_BASEFUNS(PosMGroup);
};

#define PosMGroup_of(T)							      \
class T ## _MGroup : public PosMGroup {					      \
public:									      \
  void	Initialize() 		{ SetBaseType(el_typ = &TA_ ## T); }	      \
  void 	Destroy()		{ };					      \
  TA_BASEFUNS(T ## _MGroup);						      \
}

#endif // USE_TEMPLATE_GROUPS

#endif // base_h

