/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// enviro.h

#ifndef enviro_h
#define enviro_h 1

#include <pdp/base.h>
#include <pdp/spec.h>
#include <ta_misc/datatable.h>
#include <ta_misc/script_base.h>
#include <ta_misc/random.h>
#include <ta_misc/fontspec.h>

////////////////////////////
//   Pattern/EventSpec    //
////////////////////////////

#define FOR_ITR_PAT_SPEC(pT, pel, pgrp, pitr, sT, sel, sgrp, sitr) \
for(pel = (pT*) pgrp ## FirstEl(pitr), sel = (sT*) sgrp ## FirstEl(sitr); \
    pel && sel; \
    pel = (pT*) pgrp ## NextEl(pitr), sel = (sT*) sgrp ## NextEl(sitr))

class EnviroView;
class Pattern_Group;

class PatternSpec : public BaseSubSpec {
  // ##SCOPE_Environment sub-spec for patterns within an eventspec
public:
  enum PatTypes {
    INACTIVE,			// not presented to network
    INPUT,			// input pattern
    TARGET,			// target (output) pattern
    COMPARE 			// comparison pattern (for error only)
  };

  enum PatLayer {
    FIRST,			// first layer in the network
    LAST,			// last layer in the network
    LAY_NAME,			// specify layer by name
    LAY_NUM 			// specify layer by number
  };

  enum PatFlags {		// bit-flags for the flag field on pattern
    NO_FLAGS 		= 0x00,	// no flags on pattern value, apply as normal
    TARG_FLAG 		= 0x01,	// unit's TARG flag is set
    EXT_FLAG 		= 0x02,	// unit's EXT flag is set
    TARG_EXT_FLAG 	= 0x03,	// unit's TARG and EXT flags are set
    COMP_FLAG 		= 0x04,	// unit's COMP flag is set
    COMP_TARG_FLAG	= 0x05,	// unit's COMP and TARG flags are set
    COMP_EXT_FLAG	= 0x06,	// unit's COMP and EXT flags are set
    COMP_TARG_EXT_FLAG  = 0x07,	// unit's COMP, TARG, and EXT flags are set
    TARG_VALUE 		= 0x10,	// pattern value goes to the unit targ field
    EXT_VALUE 		= 0x20,	// pattern value goes to the unit ext field
    TARG_EXT_VALUE 	= 0x30,	// pattern value goes to the unit targ & ext fields
    NO_UNIT_FLAG 	= 0x40,	// no unit flags are set, but value is set as normal
    NO_UNIT_VALUE 	= 0x80,	// don't set the unit's value, but flag as normal
    NO_APPLY 		= 0x100 // don't apply this value to unit (no flags either)
  };

  enum LayerFlags {		// how to flag the layer's external input status
    DEFAULT 		= 0x00,	// set default layer flags based on pattern type
    TARG_LAYER 		= 0x01,	// as a target layer
    EXT_LAYER 		= 0x02,	// as an external input layer
    TARG_EXT_LAYER 	= 0x03,	// as both external input and target layer
    COMP_LAYER		= 0x04,	// as a comparison layer
    COMP_TARG_LAYER	= 0x05,	// as a comparision and target layer
    COMP_EXT_LAYER	= 0x06,	// as a comparison and external input layer
    COMP_TARG_EXT_LAYER = 0x07,	// as a comparison, target, and external input layer
    NO_LAYER_FLAGS	= 0x10 	// don't set any layer flags at all
  };

  enum PatUseFlags {            // control use of the flag field
    USE_NO_FLAGS,               // no flags on pattern value, apply as normal
    USE_PATTERN_FLAGS,          // use flags from pattern
    USE_GLOBAL_FLAGS,           // use global flags from pattern spec
    USE_PAT_THEN_GLOBAL_FLAGS   // use flags from pattern if avail, else global flags
  };
  
  PatTypes	type;    	// #ENVIROVIEW Type of pattern 
  PatLayer  	to_layer;    	// #ENVIROVIEW which network layer to present pattern to
  String        layer_name;  	// #ENVIROVIEW name of layer
  int           layer_num;   	// #ENVIROVIEW number of layer
  Layer* 	layer;		// #READ_ONLY #NO_SAVE Pointer to Layer presented to 

  TypeDef*	pattern_type;	// #TYPE_Pattern type of pattern to use
  LayerFlags	layer_flags;	// #ENVIROVIEW how to flag the layer's external input status
  PatUseFlags 	use_flags;	// how to use the flags (on each pattern or global_flags)
  int		n_vals;		// number of values in pattern
  PosTDCoord    geom;		// geometry of pattern in EnviroView
  PosTDCoord    pos;		// position of pattern in EnviroView
  float		initial_val;	// Initial value for pattern values
  Random	noise;		// #ENVIROVIEW Noise added to values when applied
  String_Array  value_names;	// display names of the individual pattern values
  int_Array 	global_flags;	// these are global flags for all events (cf use_flags)
  
  virtual bool 	SetLayer(Network* net);
  // set layer pointer to point to the target layer
  virtual void 	UnSetLayer();
  // when done, don't keep pointing to it.
  virtual void	FlagLayer();
  // set layer flag to reflect the kind of input received
  virtual float& Value(Pattern* pat, int index);
  // return value at given index from pattern (order can be changed, eg GroupPatternSpec)
  virtual int&	Flag(PatUseFlags flag_type, Pattern* pat, int index);
  // return flag at given index from pattern (order can be changed, eg GroupPatternSpec)
  virtual void 	ApplyValue(Pattern* pat, Unit* uni, int index);
  // assign unit value and ext_flag based on pattern at given index
  virtual void	ApplyValue_impl(Unit* uni, float val, int flags);
  // implementation of apply value, taking just the value and the flags
  virtual void	ApplyValueWithFlags(Unit* uni, float val, int flags);
  // assign unit value and ext_flag with pattern flags
  virtual void	ApplyPattern(Pattern* pat);
  // apply given pattern to all units (layer must already be set)

  virtual Pattern* NewPattern(Event* ev, Pattern_Group* par);
  // creates a new pattern in my image in event in parent group at index
  virtual void	UpdatePattern(Event* ev, Pattern* pat);
  // updates existing pattern to current spec settings

  virtual void 	SetToLayer(Layer* lay);
  // #BUTTON set configuration of the pattern spec based on layer, and set to go to that layer
  virtual void	UpdateAllEvents();
  // #BUTTON update all events using pattern spec

  virtual Network* GetDefaultNetwork();
  // #IGNORE get default network from project

  void	UpdateAfterEdit();	// gets information off of the layer, etc
  void	Initialize();
  void 	Destroy();
  void  InitLinks();
  void	CutLinks();
  void 	Copy_(const PatternSpec& cp);
  COPY_FUNS(PatternSpec, BaseSubSpec);
  TA_BASEFUNS(PatternSpec);
};

class PatternSpec_Group : public taBase_Group {
  // ##SCOPE_Environment group of pattern specs (acts like a template for pattern groups)
public:
  TypeDef*	pat_gp_type;	// #TYPE_Pattern_Group type of pattern group to use

  // overload standard operators to have corresponding effect on patterns
  TAPtr		NewEl(int n_els=0, TypeDef* type=NULL);
  taGroup<taBase>* NewGp(int n_gps=0, TypeDef* type=NULL);
  TAPtr		New(int n_objs=0, TypeDef* type=NULL);
  bool		Remove(const char* it)	{ return taGroup_impl::Remove(it); }
  bool		Remove(TAPtr it)	{ return taGroup_impl::Remove(it); }
  bool		Remove(int i);

  virtual Pattern_Group* NewPatternGroup(Event* ev, Pattern_Group* par);
  // creates a new pattern_group in my image in event in parent group at index
  virtual void	UpdatePatternGroup(Event* ev, Pattern_Group* pg);
  // updates existing pattern group to current spec settings

  void	UpdateAfterEdit();	// call EventSpec's UpdateAllEvents()

  virtual void	LinearLayout();
  // #MENU #MENU_ON_Actions Layout PatternSpecs linearly based on EnviroView's event_layout

  void	Initialize();
  void 	Destroy();
  void	CutLinks();
  void 	Copy_(const PatternSpec_Group& cp);
  COPY_FUNS(PatternSpec_Group, taBase_Group);
  TA_BASEFUNS(PatternSpec_Group);
};

class EventSpec : public BaseSpec {
  // ##SCOPE_Environment ##MEMB_IN_GPMENU ##IMMEDIATE_UPDATE event specification 
public:
  Network*		last_net;
  // #READ_ONLY #NO_SAVE last network events applied to
  PatternSpec_Group 	patterns;  
  // #IN_GPMENU #NO_INHERIT group of pattern specs (one-to-one with patterns)

  virtual void 	ApplyPatterns(Event* ev, Network* net);
  // apply patterns to the network

  virtual void	SetLayers(Network* net); // set layers according to given net
  virtual void	UnSetLayers(); 		 // clear layer pointers (and last_net)

  int	MaxX();			// maximum X coordinate of patterns
  int	MaxY();			// maximum Y coordinate of patterns

  virtual void	NewEvent(Event* ev);
  // defines a new event in my image
  virtual void	UpdateEvent(Event* ev);
  // updates existing event to current spec settings
  virtual void	UpdateAllEvents();
  // #BUTTON update all events using this event spec

  virtual void	LinearLayout();
  // #MENU #MENU_ON_Actions Layout PatternSpecs linearly based on EnviroView's event_layout

  virtual bool	DetectOverlap();
  // determine if the patterns overlap on top of each other

  virtual void	AddToView();		// add event to view(s)
  virtual void	RemoveFromView();	// remove event from view(s)

  void		UpdateSubSpecs();
  void          UpdateSpec();

  void  UpdateAfterEdit();
  void	Initialize();
  void 	Destroy() 	{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void 	Copy(const EventSpec& cp);
  TA_BASEFUNS(EventSpec);
};

class EventSpec_SPtr : public SpecPtr<EventSpec> {
public:
  BaseSpec_MGroup*	GetSpecGroup();	// event specs go in environment
  void 	Initialize() 		{ };
  void	Destroy()		{ };
  TA_BASEFUNS(EventSpec_SPtr);
};


////////////////////////
//   Pattern/Event    //
////////////////////////

class Pattern : public taOBase {
  // ##SCOPE_Environment ##EXT_pat ##NO_TOKENS ##NO_UPDATE_AFTER generic pattern 
public:
  float_RArray 	value;  	// Values of Pattern
  int_Array   	flag;  		// Flags of Pattern

  void 	Initialize();
  void 	Destroy()		{ };
  void	InitLinks();
  void 	Copy_(const Pattern& cp);
  COPY_FUNS(Pattern, taOBase);
  TA_BASEFUNS(Pattern);
};

BaseGroup_of(Pattern);


class Event : public taNBase {
  // ##SCOPE_Environment ##EXT_evt generic event ##NO_UPDATE_AFTER
public:
  int			index;		// #NO_SAVE #READ_ONLY Index of this event within group
  Pattern_Group 	patterns;  	// #NO_SAVE_PATH_R group of patterns
  EventSpec_SPtr	spec;		// event spec

  void 		ApplyPatterns(Network* net) { spec->ApplyPatterns(this, net); }

  virtual void	GetLocalSpec();	// get event spec that is local to enviro

  virtual String GetDisplayName(); // get a name for displaying event clearly

  virtual void	AddToView();		// add event to view(s)
  virtual void	RemoveFromView();	// remove event from view(s)

  virtual void	PresentEvent(TrialProcess* trial_proc, bool new_init=false);
  // #BUTTON present this event to the given trial process, which is ReInit (or new_init) and run

  void	UpdateAfterEdit();

  void	Initialize();
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void 	Copy(const Event& cp);
  TA_BASEFUNS(Event);
};

// environment presents the following model: either a flat list of events
// (interface = EventCount() && GetEvent()) or a set of event-groups (leaf groups)
// (interface = GroupCount() && GetGroup())
// when events are generated algorithmically, the results are put in a set of event
// structures, and the same interface can be used (see ProcEnvironment below)
// other models are definable, but the standard EpochProcess will not 
// understand them.

class Event_MGroup : public PDPMGroup {
  // ##SCOPE_Environment Group of events
protected:
  void	El_SetIndex_(void* base, int idx) { ((Event*)base)->index = idx; }
public:

  virtual void	InitEvents(Environment*)	{ };
  // initialize events at the level of each event group (can be called by enviro)

  virtual int	 EventCount()		{ return leaves; }
  // #MENU #MENU_ON_Actions #USE_RVAL number of events in environment
  virtual Event* GetEvent(int i)	{ return (Event*)Leaf(i); }

  TAPtr 	New(int n_objs=0, TypeDef* typ = NULL);    // make using default spec..
  TAPtr 	NewEl(int n_els=0, TypeDef* typ = NULL);  // make using default spec..
  Event*	NewFmSpec(int n_objs, TypeDef* typ = NULL, EventSpec* es = NULL);
  // #MENU #MENU_ON_Object #TYPE_ON_el_base #UPDATE_MENUS make using given event spec
  virtual void	NewEl_impl(int old_sz, EventSpec* es=NULL);
  // #IGNORE implements event spec copying (if es is NULL, uses default)

  void	Initialize();
  void 	Destroy()		{ };
  TA_BASEFUNS(Event_MGroup);
};

////////////////////////
//   Environment      //
////////////////////////

class Environment : public WinMgr {
  // ##EXT_env ##COMPRESS generic environment
public:
  enum TextFmt {
    NAME_FIRST,			// save file with names as first column
    NAME_LAST,			// save file with names as last column
    NO_NAME			// no names at all in file..
  };

  enum DistMatFmt {		// distance matrix format
    STD_PRINT,			// standard printout format
    PRINT_NO_LABELS,		// standard printout with no event names or other labels
    GRID_LOG			// grid log format for importing into a grid log
  };

  BaseSpec_MGroup 	event_specs;	// specs for events
  Event_MGroup 		events;		// events

  virtual void	InitEvents()	{ };
  // #MENU #MENU_ON_Actions #UPDATE_MENUS initialize events for an epoch (eg, if algorithmically created)
  virtual void	UnSetLayers();
  // reset layer pointers on patterns so they are automatically recomputed when events are presented

  virtual void	UpdateAllEvents();
  // #MENU #MENU_ON_Actions update all events from their event specs

  // the flat event list model of the environment
  virtual int	EventCount()		{ return events.leaves; }
  // #MENU #MENU_ON_Actions #USE_RVAL #MENU_SEP_BEFORE number of events in environment
  virtual Event* GetEvent(int i)	{ return (Event*)events.Leaf(i); }

  // the leaf-group model of the environment
  virtual int	GroupCount();
  // #MENU #USE_RVAL number of event groups in environment
  virtual Event_MGroup* GetGroup(int i); // return the ith event group

  virtual void 	ReadText(istream& strm, EventSpec* es, TextFmt fmt = NAME_FIRST);
  // #MENU #MENU_ON_Object #MENU_SEP_BEFORE #EXT_strm_pat #UPDATE_MENUS Read text-formatted event/pattern files, including old pdp files, (using given event spec, -1=default)
  virtual void	WriteText(ostream& strm, int pat_no = -1, TextFmt fmt = NAME_FIRST);
  // #MENU #EXT_strm_pat Save enviro in text format: pat_no of -1 gives all pats, fmt for name
  virtual void	ReadBinary(istream& strm, EventSpec* es);
  // #MENU #EXT_strm_pat read event pattern information in binary format (just the numbers only)
  virtual void	WriteBinary(ostream& strm);
  // #MENU #EXT_strm_pat write event pattern information in binary format (just the numbers only)

  // the following are simple pattern generation routines

  virtual void	ReplicateEvents(int n_replicas, bool make_groups=false);
  // #MENU #MENU_ON_Generate #UPDATE_MENUS replicate existing events, optionally placing in groups

  virtual void	PermutedBinary(int pat_no, int n_on);
  // #MENU #MENU_SEP_BEFORE create permuted binary patterns of n_on 1's and rest 0's
  virtual void	PermutedBinary_MinDist(int pat_no, int n_on, float dist,
				       float_RArray::DistMetric metric=float_RArray::HAMMING,
				       bool norm=false, float tol=0.0f);
  // #MENU create permuted binary patterns with dist minimum hamming distance (or dist max_correl)
  virtual void  FlipBits(int pat_no, int n_off, int n_on);
  // #MENU flip n_off bits from 1's to 0's, and n_on bits from 0's to 1's
  virtual void  FlipBits_MinMax(int pat_no, int n_off, int n_on, float min_dist, float max_dist,
				float_RArray::DistMetric metric=float_RArray::HAMMING,
				bool norm=false, float tol=0.0f);
  // #MENU flip bits, ensuring range within min and max distances
  virtual void  FlipBits_GpMinMax(int pat_no, int n_off, int n_on, float within_min_dist,
				  float within_max_dist, float between_dist,
				  float_RArray::DistMetric metric=float_RArray::HAMMING,
				  bool norm=false, float tol=0.0f, int st_gp=0, int ed_gp=-1);
  // #MENU flip bits, ensuring within-group min and max distances, and between-group min dist

  // helper pattern-wise functions for above generation routines
  static void	PermutedBinaryPat(Pattern* pat, int n_on, float on_val=1.0f, float off_val=0.0f);
  // set pat values to permuted binary pattern of n_on on_vals and rest off_vals
  static void	FlipBitsPat(Pattern* pat, int n_off, int n_on);
  // flip n_off of the 1 bits into the 0 state, and n_on of the 0 bits to the 1 state
  static void	AddNoisePat(Pattern* pat, const Random& rnd_spec);
  // add random noise to given pattern
  virtual float LastMinDist(int n, int pat_no, float_RArray::DistMetric metric=float_RArray::HAMMING,
			    bool norm=false, float tol=0.0f);
  // returns minimum distance (or max correl) between last (n th) pattern and all previous
  virtual float LastMinMaxDist(int n, int pat_no, float& max_dist,
			       float_RArray::DistMetric metric=float_RArray::HAMMING,
			       bool norm=false, float tol=0.0f);
  // returns min and max distance between last (n th) pattern and all previous
  virtual float GpWithinMinMaxDist(Event_MGroup* gp, int n, int pat_no, float& max_dist,
				   float_RArray::DistMetric metric=float_RArray::HAMMING,
				   bool norm=false, float tol=0.0f);
  // returns min and max distance between last (n th) pattern and all previous within group
  virtual float GpLastMinMaxDist(int gp_no, Pattern* trg_pat, int pat_no, float& max_dist, 
				 float_RArray::DistMetric metric=float_RArray::HAMMING,
				 bool norm=false, float tol=0.0f, int st_gp=0);
  // returns min and max distance between patterns in all groups up to gp_no for pattern pat
  virtual float GpMinMaxDist(Event_MGroup* gp, Pattern* trg_pat, int pat_no, float& max_dist,
			     float_RArray::DistMetric metric=float_RArray::HAMMING,
			     bool norm=false, float tol=0.0f);
  // returns min and max distance between probe pattern and all in group

  virtual void	AddNoise(int pat_no, const Random& rnd_spec);
  // #MENU add random noise of specified type to the patterns

  virtual void	TransformPats(int pat_no, const SimpleMathSpec& trans);
  // #MENU #MENU_SEP_BEFORE Apply given transformation to pattern pat_no of all events

  virtual void	DistMatrix(ostream& strm, int pat_no, float_RArray::DistMetric metric=float_RArray::HAMMING,
			   bool norm=false, float tol=0.0f, DistMatFmt format=STD_PRINT, int precision = -1);
  // #MENU #EXT_strm_dstmat #MENU_SEP_BEFORE output distance matrix for events based on pattern pat_no
  virtual void	DistMatrixGrid(GridLog* disp_log, int pat_no, float_RArray::DistMetric metric=float_RArray::HAMMING,
			   bool norm=false, float tol=0.0f);
  // #MENU #NULL_OK output to grid log distance matrix for events based on pattern pat_no
  virtual void	DistArray(float_RArray& dist_ary, int pat_no,
			  float_RArray::DistMetric metric=float_RArray::HAMMING,
			  bool norm=false, float tol=0.0f);
  // get distance matrix as an array for events based on pattern pat_no
  virtual void	GpDistArray(float_RArray& within_dist_ary, float_RArray& between_dist_ary, int pat_no,
			    float_RArray::DistMetric metric=float_RArray::HAMMING,
			    bool norm=false, float tol=0.0f);
  // get within group and between group distance matricies as arrays for events based on pattern pat_no

  virtual void	CmpDistMatrix(ostream& strm, int pat_no, Environment* cmp_env, int cmp_pat_no,
			      float_RArray::DistMetric metric=float_RArray::HAMMING,
			      bool norm=false, float tol=0.0f, DistMatFmt format=STD_PRINT);
  // #MENU #EXT_strm_dstmat comparative distance array between two environments
  virtual void	CmpDistArray(float_RArray& dist_ary, int pat_no, Environment* cmp_env, int cmp_pat_no,
			     float_RArray::DistMetric metric=float_RArray::HAMMING,
			     bool norm=false, float tol=0.0f);
  // comparative distance array between two environments

  virtual void	CmpDistArrayPat(float_RArray& dist_ary, Pattern* trg_pat,
				int cmp_pat_no, float_RArray::DistMetric metric=float_RArray::HAMMING,
				bool norm=false, float tol=0.0f);
  // compute comparative distance array, one pattern against this environment

  virtual void	ClusterPlot(GraphLog* disp_log, int pat_no, bool graphic=true,
			    float_RArray::DistMetric metric=float_RArray::EUCLIDIAN,
			    bool norm=false, float tol=0.0f);
  // #MENU #NULL_OK produce a cluster plot (in graph log, NULL = make a new one) of the given pat_no across events
  virtual void	EnvToGrid(GridLog* disp_log, int pat_no, int ev_x=-1, int ev_y=-1, int pt_x=-1, int pt_y=-1);
  /* #MENU #MENU_SEP_BEFORE #NULL_OK send environment to grid log, with given layout 
     (-1 = default, ev = event layout, pt = pattern layout) */

  void	ReplacePointersHook(TAPtr old);

  TypeDef*	GetDefaultView()	{ return &TA_EnviroView; }

  void	Initialize();
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void 	Copy(const Environment& cp);
  TA_BASEFUNS(Environment);
};

class Environment_MGroup : public PDPMGroup {
  // Menu group of environments
public:
  void  Initialize()            { SetBaseType(&TA_Environment); }
  void  Destroy()               { };
  TA_BASEFUNS(Environment_MGroup);
};

class Environment_List : public taList<Environment> {
  // ##NO_TOKENS ##NO_UPDATE_AFTER a simple list of environments
public:
  void  Initialize()            { SetBaseType(&TA_Environment); }
  void  Destroy()               { };
  TA_BASEFUNS(Environment_List);
};

class EnvEditor;
class ColorScaleSpec;

class EnviroView : public PDPView { 
  // ##SCOPE_Environment Graphical view of an Environment
public:  
  enum  EventLayout {
    HORIZONTAL, //  layout events horizontally
    VERTICAL 	// layout events vertically
  };

  enum ValDispMode {		// ways of displaying values
    COLOR,			// color indicates value
    AREA,			// area indicates value
    LINEAR 			// linear size of square side indicates value
  };

  enum ValTextMode {		// how to display text for values
    NONE,			// no text (default)
    VALUES,			// values only
    LABELS,			// labels only
    BOTH 			// both unit values and labels
  };

  EventLayout	  event_layout;	// layout of the events when mulitple are displayed
  bool		  auto_scale;	// whether to update the min/max range automatically
  ValDispMode	  val_disp_mode; // method of displaying values
  ColorScaleSpec* colorspec;	// color scale to use
  ValTextMode	  val_text;	// how to display the text values
  bool		  no_border;	// do not display a border around the values
  FontSpec	  view_font;	// the font to use for the labels in the display
  FontSpec	  value_font;	// the font to use for the values

  Environment* 	  env;			// #HIDDEN #NO_SAVE
  MinMax          scale_range;		// #HIDDEN range of scalebar
  EnvEditor*	  editor;		// #READ_ONLY #NO_SAVE environment editor for this one
  Event_MGroup    events_displayed;	// #HIDDEN #NO_SAVE
  BaseSpec_MGroup specs_displayed;	// #HIDDEN #NO_SAVE

  virtual void	SetEventSpec(EventSpec* es);
  // #MENU #MENU_ON_Actions Set event spec for selected events

  virtual void	SelectEvents(int start=0, int n_events=-1);
  // #MENU #MENU_SEP_BEFORE select events to view starting with given event continuing until n (-1 = to end)
  virtual void	DeselectEvents(int start=0, int n_events=-1);
  // #MENU deselect events to view starting with given event continuing until n (-1 = to end)

  void 		InitDisplay();
  void		UpdateDisplay(TAPtr updtr=NULL);

  void 		GetBodyRep();	// #IGNORE
  void          CloseWindow();
  ivGlyph*	GetPrintData();	// #IGNORE

  virtual Network* GetDefaultNetwork();
  // #IGNORE get default network from project

  void	UpdateAfterEdit();
  void 	Initialize();
  void 	InitLinks();
  void  CutLinks();
  void 	Destroy();
  void	Copy_(const EnviroView& cp);
  COPY_FUNS(EnviroView, PDPView);
  TA_BASEFUNS(EnviroView);
};


#endif // enviro_h
