/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// sched_proc.cc


#include <pdp/sched_proc.h>
#include <pdp/procs_extra.h>
#include <ctype.h>
#include <ta/taiv_dialog.h>

void StepParams::Initialize() {
  owner = NULL;
  proc = NULL;
  n = 1;
}

void StepParams::InitLinks() {
  taBase::InitLinks();
  taBase::SetPointer((TAPtr*)&proc, owner);
}


void StepParams::CutLinks() {
  owner = NULL;
  taBase::DelPointer((TAPtr*)&proc);
  taBase::CutLinks();
}

TAPtr 	StepParams::GetOwner() const	{ return (taBase*)owner; }
TAPtr	StepParams::GetOwner(TypeDef* tp) const { return taBase::GetOwner(tp); }
TAPtr 	StepParams::SetOwner(TAPtr ta)	{ owner = (SchedProcess*)ta; return ta; }

void StepParams::Copy_(const StepParams& cp) {
  n = cp.n;
}


//////////////////////////
// 	SchedProcess	//
//////////////////////////

// flag that determines if a step function call was made 
static bool did_step_call = false;

void SchedProcess::Initialize() {
  re_init=true;
  sub_proc_type = NULL;
  can_stop = true;
  cntr = NULL;
  super_proc = NULL;
  sub_proc = NULL;
  step_flag = false;
  log_loop = false;
  log_counter = false;
  loop_stats.SetBaseType(&TA_Stat);
  final_stats.SetBaseType(&TA_Stat);
  init_procs.SetBaseType(&TA_Process);
  loop_procs.SetBaseType(&TA_Process);
  final_procs.SetBaseType(&TA_Process);
  displays.SetBaseType(&TA_PDPView);
  logs.SetBaseType(&TA_PDPLog);
}

void SchedProcess::InitLinks() {
  Process::InitLinks();
  TypeDef* td = GetTypeDef();
  int i;
  for(i=TA_SchedProcess.members.size;i<td->members.size;i++){
    MemberDef* md = td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_Counter)){
      cntr = (Counter *) md->GetOff(this);
      cntr->name = md->name;
      break;
    }
  }
  taBase::Own(step, this);
  taBase::Own(loop_stats, this);
  taBase::Own(final_stats, this);
  taBase::Own(init_procs, this);
  taBase::Own(loop_procs, this);
  taBase::Own(final_procs, this);
  taBase::Own(displays, this);
  taBase::Own(logs, this);
  taBase::Own(cntr_items, this);

  // set name to be shorter version of original (no Process or type names)
  if(name.length() == 0)
    return;
  String nw_nm = name.before((char)'_',-1);
  if(nw_nm.length() == 0)
    return;
  if(nw_nm.contains("Process"))
    nw_nm.gsub("Process","");
  else if(nw_nm.contains("Proc"))
    nw_nm.gsub("Proc", "");
  else {
    int pos = nw_nm.length()-1;
    int c = nw_nm[pos];
    while(!isupper(c) && (pos > 0)) c = nw_nm[--pos];
    if(pos > 0)
      nw_nm = nw_nm.from(pos);
  }
  name = nw_nm + name.from('_',-1);
}

void SchedProcess::Destroy() {
  CutLinks();
}

void SchedProcess::CutLinks() {
  RemoveFromLogs();		// logs have a pointer to us in them too
  RemoveFromDisplays();		// displays have a pointer to us in them too
  logs.CutLinks();
  displays.CutLinks();
  loop_stats.CutLinks();
  final_stats.CutLinks();
  init_procs.CutLinks();
  loop_procs.CutLinks();
  final_procs.CutLinks();
  super_proc = NULL;
  cntr = NULL;
  step.CutLinks();
  if(sub_proc != NULL) {
    Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
    if(gp != NULL)
      gp->RemoveLeaf(sub_proc);
  }
  taBase::DelPointer((TAPtr*)&sub_proc);
  Process::CutLinks();
}

void SchedProcess::Copy_(const SchedProcess& cp) {
  step_flag = cp.step_flag;
  cntr_items = cp.cntr_items;
  re_init = cp.re_init;
  can_stop = cp.can_stop;
  sub_proc_type = cp.sub_proc_type;
  step = cp.step;
  loop_stats = cp.loop_stats;
  final_stats = cp.final_stats;
  init_procs = cp.init_procs;
  loop_procs = cp.loop_procs;
  final_procs = cp.final_procs;
  displays.BorrowUnique(cp.displays);	// don't replicate because we don't own them...
  logs.BorrowUnique(cp.logs);
  log_loop = cp.log_loop;
  log_counter = cp.log_counter;

  super_proc = cp.super_proc;
  taBase::SetPointer((TAPtr*)&sub_proc, cp.sub_proc);
  // set our pointer, but the sub_proc doesn't think it belongs under us!

  // this is only for duplicating the entire hierarchy -- not general purpose
  // change duplicate function to do this?
//    if(cp.sub_proc == NULL)
//      taBase::SetPointer((TAPtr*)&sub_proc, NULL);
//    else {
//      CreateSubProcs(false);	// make sure we have the right type..
//      if(sub_proc != NULL)
//        sub_proc->UnSafeCopy(cp.sub_proc); // keep on copying..
//    }
  UpdateLogUpdaters();
  UpdateDisplayUpdaters();
}

SchedProcess* SchedProcess::FindSubProc(TypeDef* td) {
  SchedProcess* sp = sub_proc;
  while((sp != NULL) && !(sp->InheritsFrom(td)))
    sp = sp->sub_proc;
  return sp;
}

SchedProcess* SchedProcess::FindSuperProc(TypeDef* td) {
  SchedProcess* sp = super_proc;
  while((sp != NULL) && !(sp->InheritsFrom(td)))
    sp = sp->super_proc;
  return sp;
}

SchedProcess* SchedProcess::FindProcOfType(TypeDef* td) {
  if(InheritsFrom(td))
    return this;
  else {
    SchedProcess* rval = FindSuperProc(td);
    if(rval == NULL)
      rval = FindSubProc(td);
    return rval;
  }
}

SchedProcess* SchedProcess::GetTopProc() {
  SchedProcess* sp = this;
  while(sp->super_proc != NULL) sp = sp->super_proc;
  return sp;
}

void SchedProcess::RemoveFromLogs() {
  PDPLog* lg;
  taLeafItr i;
  FOR_ITR_EL(PDPLog, lg, logs., i) {
    if(lg->log_proc.Remove(this)) {
      lg->SyncLogViewUpdaters();
    }
  }
}

void SchedProcess::RemoveFromDisplays() {
  WinView* wv;
  taLeafItr i;
  FOR_ITR_EL(WinView, wv, displays., i)
    wv->updaters.Remove(this);
}

void SchedProcess::UpdateLogUpdaters() {
  // make sure the logs have us in them too...
  PDPLog* plog;
  taLeafItr i;
  FOR_ITR_EL(PDPLog, plog, logs., i) {
    if(plog->log_proc.LinkUnique(this))
      plog->SyncLogViewUpdaters();
    if(plog->cur_proc != this) {
      if(taMisc::is_loading || taMisc::is_duplicating)
	plog->cur_proc = this;
      else {
	GenLogData();
	plog->NewHead(log_data, this);
      }
    }
  }
}

void SchedProcess::UpdateDisplayUpdaters() {
  WinView* wv;
  taLeafItr i;
  FOR_ITR_EL(WinView, wv, displays., i)
    wv->updaters.LinkUnique(this);
}

void SchedProcess::InitMyLogs() {
  PDPLog* lg;
  taLeafItr i;
  FOR_ITR_EL(PDPLog, lg, logs., i)
    lg->Clear();
}

void SchedProcess::InitNetwork() {
  if(network == NULL)
    return;
  network->InitWtState();
}

void SchedProcess::InitAll() {
  Init_force();
  InitNetwork();
  InitAllLogs();
}

void SchedProcess::Run_gui() {
  NotifyNetObjects();
  Process::Run_gui();
}

void SchedProcess::Step_gui() {
  NotifyNetObjects();
  Process::Step_gui();
}

void SchedProcess::NotifyNetObjects() {
  if(!taivMisc::iv_active)    return;
  int i;
  for(i=taivMisc::active_edits.size-1; i>=0; i--) {
    taivEditDialog* dlg = taivMisc::active_edits.FastEl(i);
    if((dlg->state != taivDialog::ACTIVE) || (dlg->cur_base == NULL) ||
       (dlg->typ == NULL))
      continue;
    if(dlg->typ->InheritsFrom(TA_Network) || dlg->typ->InheritsFrom(TA_Layer) ||
       dlg->typ->InheritsFrom(TA_Projection) || dlg->typ->InheritsFrom(TA_Unit) ||
       dlg->typ->InheritsFrom(TA_Connection) || dlg->typ->InheritsFrom(TA_Con_Group))
      dlg->SetRevert();
  }
}

void SchedProcess::UpdateAfterEdit() {
  Process::UpdateAfterEdit();

  taLeafItr i;
  Process* pr;
  FOR_ITR_EL(Process, pr, final_stats., i)
    pr->CopyPNEPtrs(network, environment);
  FOR_ITR_EL(Process, pr, loop_stats., i)
    pr->CopyPNEPtrs(network, environment);
  FOR_ITR_EL(Process, pr, init_procs., i) {
    if(pr->GetOwner(GetTypeDef()) == this) { // only if we own the process
      pr->CopyPNEPtrs(network, environment);
      pr->UpdateAfterEdit();
  }
  }
  FOR_ITR_EL(Process, pr, loop_procs., i) {
    if(pr->GetOwner(GetTypeDef()) == this) { // only if we own the process
      pr->CopyPNEPtrs(network, environment);
      pr->UpdateAfterEdit();
    }
  }
  FOR_ITR_EL(Process, pr, final_procs., i) {
    if(pr->GetOwner(GetTypeDef()) == this) { // only if we own the process
      pr->CopyPNEPtrs(network, environment);
      pr->UpdateAfterEdit();
    }
  }

  GetCntrDataItems();

  CreateSubProcs(false);
  UpdateLogUpdaters();
  UpdateDisplayUpdaters();
}

void SchedProcess::CreateSubProcs(bool update) {
  if(sub_proc_type == NULL)
    return;

  if((sub_proc == NULL) || (sub_proc->GetTypeDef() != sub_proc_type)) {
    Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
    if(gp == NULL)
      return;

    if(sub_proc != NULL) {
      gp->RemoveLeaf(sub_proc);
      taBase::DelPointer((TAPtr*)&sub_proc);
    }

    sub_proc = (SchedProcess*)taBase::MakeToken(sub_proc_type);
    taBase::Ref(sub_proc);	// give it an extra reference for us
    sub_proc->super_proc = this;
    sub_proc->CopyPNEPtrs(network, environment);
    gp->Add(sub_proc); // InitLinks called with super_proc in place
    taBase::SetPointer((TAPtr*)&(step.proc), sub_proc);	// default is to step lower guy.
    if(!taMisc::is_loading && (network != NULL) &&
       (sub_proc->InheritsFrom(TA_TrialProcess) ||
	sub_proc->InheritsFrom(TA_SettleProcess)))
    {
      // auto add updater to network 
      if(network->views.size > 0)
	((NetView*)network->views.FastEl(0))->AddUpdater(sub_proc);
    }
  }

  if(sub_proc != NULL) {
    sub_proc->super_proc = this;
    sub_proc->CopyPNEPtrs(network, environment);
    sub_proc->UpdateAfterEdit();
  }

  if(update)
    UpdateAfterEdit();
}

void SchedProcess::UnLinkSubProc() {
  if(sub_proc != NULL)
    sub_proc->super_proc = NULL;
  sub_proc_type = NULL;
  taBase::SetPointer((TAPtr*)&(sub_proc), NULL);
}

void SchedProcess::UnLinkFmSuperProc() {
  if(super_proc != NULL)
    super_proc->UnLinkSubProc();
  super_proc = NULL;
}

void SchedProcess::LinkSubProc(SchedProcess* proc) {
  if(proc == NULL) {
    UnLinkSubProc();
    return;
  }
  // warning -- does nothing to preserve sub_proc's current super-proc!
  sub_proc_type = proc->GetTypeDef();
  taBase::SetPointer((TAPtr*)&(sub_proc), proc);
  sub_proc->super_proc = this;
  sub_proc->CopyPNEPtrs(network, environment);
}

void SchedProcess::LinkSuperProc(SchedProcess* proc) {
  if(proc == NULL) {
    UnLinkFmSuperProc();
    return;
  }
  // warning: does nothing to preserve super-proc's current sub-proc!
  super_proc = proc;
  super_proc->CopyPNEPtrs(network, environment);
  super_proc->LinkSubProc(this);
}

void SchedProcess::GetAggsFmSuperProc() {
  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, super_proc->loop_stats., i) {
    if(st->time_agg.from == NULL) continue; // ignore non-aggs
    Stat* nag = (Stat*)st->Clone(); // clone original one
    loop_stats.Add(nag);  	// i get it 
    taBase::SetPointer((TAPtr*)&(st->time_agg.from), nag); // super now aggs from me
    nag->UpdateAfterEdit();	// update everything on the new aggregator..
  }
}

void SchedProcess::MoveToSubGp(const char* gp_name) {
  Process_MGroup* own_gp = (Process_MGroup*)GetOwner(&TA_Process_MGroup);
  Process_MGroup* nw_gp = (Process_MGroup*)own_gp->NewGp(1);
  nw_gp->name = gp_name;
  SchedProcess* cp = this;
  while(cp != NULL) {
    nw_gp->Transfer(cp);
    cp = cp->sub_proc;
  }
  winbMisc::DelayedMenuUpdate(own_gp);
}

void SchedProcess::ChangeNameSuffix(const char* new_name_sufx) {
  SchedProcess* cp = this;
  while(cp != NULL) {
    if(cp->name.contains('_'))
      cp->name = cp->name.through('_') + new_name_sufx;
    cp = cp->sub_proc;
  }
  winbMisc::DelayedMenuUpdate(this);
}

void SchedProcess::SetAggsAfterRmvSubProc() {
  int lf;
  for(lf = loop_stats.leaves-1; lf>=0; lf--) {
    Stat* st = (Stat*)loop_stats.Leaf(lf);
    if(st->time_agg.from == NULL) continue; // ignore non-aggs
    Stat* prv_frm = st->time_agg.from;	// this is stat on old sub proc
    if(prv_frm->time_agg.from == NULL)	{ // this is not itself an agg
      loop_stats.RemoveLeaf(lf); // what i'm aggregating is dissappearing, so die!
      continue;
    }
    Stat* nw_frm = prv_frm->time_agg.from; // this is who i now want to agg from
    taBase::SetPointer((TAPtr*)&(st->time_agg.from), nw_frm);
    nw_frm->UpdateAfterEdit();	// update everything on the new aggregator..
  }
}

void SchedProcess::AddSuperProc(TypeDef* proc_type) {
  Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
  if((gp == NULL) || (proc_type == NULL) || !proc_type->InheritsFrom(&TA_SchedProcess))
    return;

  SchedProcess* prv_super = super_proc;	// cache orig
  SchedProcess* nw_proc = (SchedProcess*)taBase::MakeToken(proc_type);
  if(nw_proc == NULL) return;
  LinkSuperProc(nw_proc);
  int myidx = gp->FindEl(this);
  gp->Insert(nw_proc, myidx);	// initlinks called here, put in my place!
  if(prv_super != NULL) {
    prv_super->LinkSubProc(nw_proc); // insert into hierarchy
    super_proc->GetAggsFmSuperProc(); // new super gets aggs from its super!
  }
  else {
    // go through my stats and create aggregates of things that are already aggregators
    Stat* st;
    taLeafItr i;
    FOR_ITR_EL(Stat, st, loop_stats., i) {
      if(st->time_agg.from == NULL) continue; // ignore non-aggs
      st->CreateAggregates();
    }
  }
  winbMisc::DelayedMenuUpdate(gp);
}

void SchedProcess::AddSubProc(TypeDef* proc_type) {
  Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
  if((gp == NULL) || (proc_type == NULL) || !proc_type->InheritsFrom(&TA_SchedProcess))
    return;

  SchedProcess* prv_sub = sub_proc;	// cache orig
  SchedProcess* nw_proc = (SchedProcess*)taBase::MakeToken(proc_type);
  if(nw_proc == NULL) return;
  LinkSubProc(nw_proc);
  int myidx = gp->FindEl(this);
  gp->Insert(nw_proc, myidx+1);	// initlinks called here, put in after me
  if(prv_sub != NULL) {
    nw_proc->LinkSubProc(prv_sub); 	// insert into hierarchy
  }
  sub_proc->GetAggsFmSuperProc(); // sub gets aggs from me!
  winbMisc::DelayedMenuUpdate(gp);
}

void SchedProcess::RemoveSuperProc() {
  Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
  if((gp == NULL) || (super_proc == NULL))
    return;

  SchedProcess* prv_super = super_proc;	// cache orig
  SchedProcess* super_super = prv_super->super_proc;
  UnLinkFmSuperProc();		// disconnect me from that one!
  if(super_super != NULL) {
    super_super->LinkSubProc(this); // i am now its sub proc
    super_super->SetAggsAfterRmvSubProc(); // the super-super proc now has a new sub-proc
  }
  gp->RemoveEl(prv_super);	// get rid of it!
  winbMisc::DelayedMenuUpdate(gp);
}

void SchedProcess::RemoveSubProc() {
  Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
  if((gp == NULL) || (sub_proc == NULL))
    return;

  SchedProcess* prv_sub = sub_proc;	// cache orig
  SchedProcess* sub_sub = prv_sub->sub_proc;
  prv_sub->UnLinkSubProc();	// disconnect sub-sub from sub process, otherwise it will take it all down with it
  UnLinkSubProc();		// disconnect me from that one!
  if(sub_sub != NULL) {
    LinkSubProc(sub_sub); 	// i am now its super proc
    SetAggsAfterRmvSubProc(); 	// i need to point to new sub_sub proc aggs
  }
  gp->RemoveEl(prv_sub);	// get rid of it!
  winbMisc::DelayedMenuUpdate(gp);
}

void SchedProcess::ReplacePointersHook(TAPtr old) {
  // go through entire hierarchy and replace any pointer to old guy with me

  SchedProcess* spold = (SchedProcess*)old;
  // make sure this guy doesn't delete everything below it!
  spold->sub_proc_type = NULL;
  taBase::SetPointer((TAPtr*)&(spold->sub_proc), NULL);
  // and disconnect from above
  spold->super_proc = NULL;

  // and now connect us up with our new folks:
  if(sub_proc != NULL) {
    sub_proc->super_proc = this;
  }
  if(super_proc != NULL) {	// sp is my super-proc
    super_proc->LinkSubProc(this);
  }

  Process_MGroup* gp = GET_MY_OWNER(Process_MGroup);
  if(gp == NULL) return;
  Process_MGroup* prv_gp = gp;
  while(gp != NULL) {		// get highest level of process mgroup!
    prv_gp = gp;
    gp = GET_OWNER(gp, Process_MGroup);
  }
  
  String old_path = old->GetPath();

  // now go through the procs and find any pointers to old!
  Process* proc;
  taLeafItr pi;
  FOR_ITR_EL(Process, proc, prv_gp->, pi) {
    if(!proc->InheritsFrom(&TA_SchedProcess)) continue;
    SchedProcess* sp = (SchedProcess*)proc;
    Stat* st;
    taLeafItr i;
    FOR_ITR_EL(Stat, st, sp->final_stats., i) {
      Stat* fm = st->time_agg.from;
      if(fm == NULL) continue;
      if(GET_OWNER(fm, SchedProcess) == spold) { // points to stat owned by old guy
	String nw_pth = fm->GetPath();
	nw_pth = nw_pth.after(old_path);
	Stat* myst = (Stat*)FindFromPath(nw_pth);
	if(myst != NULL) {
	  taBase::SetPointer((TAPtr*)&(st->time_agg.from), myst); // update pointer
	}
      }
    }
    FOR_ITR_EL(Stat, st, sp->loop_stats., i) {
      Stat* fm = st->time_agg.from;
      if(fm == NULL) continue;
      if(GET_OWNER(fm, SchedProcess) == spold) { // points to stat owned by old guy
	String nw_pth = fm->GetPath();
	nw_pth = nw_pth.after(old_path);
	Stat* myst = (Stat*)FindFromPath(nw_pth);
	if(myst != NULL) {
	  taBase::SetPointer((TAPtr*)&(st->time_agg.from), myst); // update pointer
	}
      }
    }
    // replace any links pointing to me!
    
    int ri;
    for(ri=0;ri<sp->init_procs.size; ri++) { // only works for top-level..
      Process* prc = (Process*)sp->init_procs.El(ri);
      if(prc == spold) sp->init_procs.ReplaceLink(ri, this);
    }
    for(ri=0;ri<sp->loop_procs.size; ri++) { // only works for top-level..
      Process* prc = (Process*)sp->loop_procs.El(ri);
      if(prc == spold) sp->loop_procs.ReplaceLink(ri, this);
    }
    for(ri=0;ri<sp->final_procs.size; ri++) { // only works for top-level..
      Process* prc = (Process*)sp->final_procs.El(ri);
      if(prc == spold) sp->final_procs.ReplaceLink(ri, this);
    }
    sp->UpdateAfterEdit();	// update this guy to get any new pointers!
  }    
  winbMisc::DelayedMenuUpdate(prv_gp);
}

void SchedProcess::InitProcs() {
  taLeafItr i;
  Process* p;
  FOR_ITR_EL(Process, p, init_procs., i) {
    if(p->mod.flag)
      p->Run();
  }
}

void SchedProcess::Init_impl() {
  SetDefaultPNEPtrs();
  if(cntr != NULL)
    *cntr = 0;
  if(taMisc::iv_active)
    ctrl_panel.Revert();
  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, final_stats., i)
    st->Init();
  FOR_ITR_EL(Stat, st, loop_stats., i)
    st->Init();
  Process* proc;
  FOR_ITR_EL(Process, proc, init_procs., i) {
    if(!proc->InheritsFrom(TA_SchedProcess))
      proc->Init();
  }
  FOR_ITR_EL(Process, proc, loop_procs., i) {
    if(!proc->InheritsFrom(TA_SchedProcess))
      proc->Init();
  }
  FOR_ITR_EL(Process, proc, final_procs., i) {
    if(!proc->InheritsFrom(TA_SchedProcess))
      proc->Init();
  }
}

void SchedProcess::Init_step_flags() {
  SchedProcess* sp = this;
  while(sp != NULL) {		// reset the step flags
    sp->step_flag = false;
    sp = sp->sub_proc;
  }
  did_step_call = false;
}

void SchedProcess::NewInit() {
  running = false;
  iv_running = false;
  if(!CheckAllTypes())		// do this before any actions in Init_force
    return;
  rnd_seed.NewSeed();
  Init_force();
  rnd_seed.OldSeed();		// then restore seed again to start over from same place
}

void SchedProcess::ReInit() {
  running = false;
  iv_running = false;
  if(!CheckAllTypes())		// do this before any actions in Init_force
    return;
  rnd_seed.OldSeed();
  Init_force();
  rnd_seed.OldSeed();		// then restore seed again to start over from same place
}

void SchedProcess::Init() {
  rnd_seed.GetCurrent();
  Init_impl();
  SetReInit(false);
  if(sub_proc != NULL)		// set sub-procs to be re-inited when run
    sub_proc->Init_flag();
  taLeafItr i;
  SchedProcess* proc;
  FOR_ITR_EL(SchedProcess, proc, init_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
  FOR_ITR_EL(SchedProcess, proc, loop_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
  FOR_ITR_EL(SchedProcess, proc, final_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
}

void SchedProcess::Init_flag() {
  SetReInit(true);
  if(sub_proc != NULL)		// set sub-procs to be re-inited when run
    sub_proc->Init_flag();
  taLeafItr i;
  SchedProcess* proc;
  FOR_ITR_EL(SchedProcess, proc, init_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
  FOR_ITR_EL(SchedProcess, proc, loop_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
  FOR_ITR_EL(SchedProcess, proc, final_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_flag();
  }
}

void SchedProcess::Init_force() {
  SetReInit(true);		// make sure init flag is set at this point
  Init_impl();
  if(sub_proc != NULL)		// force re-init of sub-procs
    sub_proc->Init_force();
  taLeafItr i;
  SchedProcess* proc;
  FOR_ITR_EL(SchedProcess, proc, init_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_force();
  }
  FOR_ITR_EL(SchedProcess, proc, loop_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_force();
  }
  FOR_ITR_EL(SchedProcess, proc, final_procs., i) {
    if(proc->InheritsFrom(TA_SchedProcess))
      proc->Init_force();
  }
  Init_step_flags();
}

bool SchedProcess::CheckAllTypes() {
  CheckResetCache();		// reset the cache first
  return CheckAllTypes_impl();
}

bool SchedProcess::CheckAllTypes_impl() {
  if(!CheckNetwork())
    return false;

  // go depth-first on type checking to get most sensitive processes first
  // then get more generic things like stats, which will already have been
  // dealt with by inheritance
  if(sub_proc != NULL) {
    if(!sub_proc->CheckAllTypes_impl())
      return false;
  }

  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, loop_stats., i) {
    if(!st->CheckNetwork())
      return false;
  }
  FOR_ITR_EL(Stat, st, final_stats., i) {
    if(!st->CheckNetwork())
      return false;
  }
  Process* proc;
  FOR_ITR_EL(Process, proc, init_procs., i) {
    if(!proc->CheckNetwork())
      return false;
  }
  FOR_ITR_EL(Process, proc, loop_procs., i) {
    if(!proc->CheckNetwork())
      return false;
  }
  FOR_ITR_EL(Process, proc, final_procs., i) {
    if(!proc->CheckNetwork())
      return false;
  }

  return true;
}

void SchedProcess::Step() {
  SchedProcess* sp = this;
  // based on the sub_proc, not the proc itself.
  // set step flags of all containing procs
  while((sp != NULL) && (this != step.proc) && (sp->sub_proc != step.proc)) {
    sp->step_flag = true;
    sp = sp->sub_proc;
  }

  if(sp == NULL) return;

  sp->step_flag = true;		// and the step proc itself
  sp->step.n = step.n;		// copy step n down here, where it is actually used
  did_step_call = true;		// actually did a step call at some point..

#ifdef HAVE_VOLATILE
  volatile SchedProcess* ths = this; // for setjmp's
#else
  SchedProcess* ths = this; // for setjmp's
#endif
  if(setjmp(pdpMisc::proc_begin) == 0)
    ths->Run();
  did_step_call = false;
}

void SchedProcess::GoTo(int goto_val) {
  Init_force();			// first initialize (gets all sub-guys inited too)
  if(cntr != NULL)
    cntr->val = goto_val;
  SetReInit(false);		// don't init again!
  UpdateState();		// then update state for new val
}

void SchedProcess::C_Code() {
  bool critval;

  if(re_init) {			// if its time to re-initialize, then do it
    Init();			// this sets the counters to zero, etc.
    InitProcs();		// this runs any initialization processes
  }

  do {
    Loop();			// user defined code goes here
    UpdateCounters();		// update the counters (before logging)
    LoopProcs();		// check/run loop procs (use mod based on counter)
    LoopStats();		// udpate in-loop stats
    if(log_loop)
      UpdateLogs();		// generate log output and update logs
    UpdateState();		// update process state vars

    critval = Crit();		// check if stopping criterion was reached
    if(!critval) {		// if at critera, going to quit anyway, so don't
      StepCheck();		// check for stepping
      StopCheck();		// check for stopping
    }
  }
  while(!critval);		// loop until we reach criterion (e.g. ctr > max)

  Final();			// user defined code at end of loop
  FinalProcs();			// call the final procs
  FinalStats();			// final statistics after looping
  if(!log_loop)
    UpdateLogs();		// if not loggin in loop, logging at end
  UpdateDisplays();		// update displays after the loop
  SetReInit(true);		// made it through the loop, so Init next time
}

void SchedProcess::Loop() {
  if(sub_proc != NULL) sub_proc->Run();
}

void SchedProcess::UpdateCounters() {
  if(cntr) cntr->Inc();
}

void SchedProcess::LoopStats() {
  int c = GetCounter();
  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, loop_stats., i) {
    if(st->mod.flag && (((c - st->mod.off) % (st->mod.m)) == 0)){
      if((st->time_agg.from == NULL) && st->loop_init) st->Init();
      st->Run();
    }
  }
}

void SchedProcess::UpdateState() {
  if(taMisc::iv_active)
    ctrl_panel.Revert();
}

void SchedProcess::LoopProcs() {
  int c = GetCounter();
  taLeafItr i;
  Process* p;
  FOR_ITR_EL(Process, p, loop_procs., i) {
    if(p->mod.flag && (((c - p->mod.off) % (p->mod.m)) == 0))
      p->Run();
  }
}

void SchedProcess::UpdateDisplays() {
  if(!taMisc::iv_active)
    return;

  ctrl_panel.Revert();

  WinView* wv;
  taLeafItr i;
  FOR_ITR_EL(WinView, wv, displays., i) {
    wv->UpdateDisplay(this);
  }
}
  
void SchedProcess::StepCheck() {
  if((step_flag == false) || (did_step_call == false))
    return;

  if(step.n <= 0) step.n = 1;

  if((sub_proc == NULL) || ((GetCounter() % step.n) == 0)) {
    step_flag = false;		// its a toggle
    longjmp(pdpMisc::proc_begin,2);	// bail out
  }
}
  
void SchedProcess::StopCheck() {
  if(!can_stop || !taMisc::iv_active)
    return;

  if(HandleEvents() == Process::STOP) longjmp(pdpMisc::proc_begin,1);
}


bool SchedProcess::Crit() {
  if((cntr != NULL) && (cntr->Crit()))
    return true;		// reached criterion by counter, bail

  int lc = GetCounter();	// loop counter
  int fc = 0;			// final counter
  if(super_proc != NULL)
    fc = super_proc->GetCounter() + 1; // add 1 to simulate increment of parent log..

  taLeafItr i;
  Stat* st;
  FOR_ITR_EL(Stat, st, final_stats., i) {
    if(st->mod.flag && (((fc - st->mod.off) % (st->mod.m)) == 0) && st->Crit())
      return true;
  }
  FOR_ITR_EL(Stat, st, loop_stats., i) {
    if(st->mod.flag && (((lc - st->mod.off) % (st->mod.m)) == 0) && st->Crit())
      return true;
  }
  return false;
}

void SchedProcess::FinalStats() {
  int c = 0;
  if(super_proc != NULL)
    c = super_proc->GetCounter() + 1; // add 1 to simulate increment of parent log..
  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, final_stats., i) {
    if(st->mod.flag && (((c - st->mod.off) % (st->mod.m)) == 0))
      st->Run();
  }
}

void SchedProcess::FinalProcs() {
  int c = 0;
  if(super_proc != NULL)
    c = super_proc->GetCounter() + 1; // add 1 to simulate increment of parent log..
  taLeafItr i;
  Process* p;
  FOR_ITR_EL(Process, p, final_procs., i) {
    if(p->mod.flag && (((c - p->mod.off) % (p->mod.m)) == 0))
      p->Run();
  }
}

void SchedProcess::GetCntrDataItems() {
  if((cntr_items.size >= 1) || (cntr == NULL))
    return;
  DataItem* it = (DataItem*)cntr_items.New(1,&TA_DataItem);
  it->SetNarrowName(cntr->name);
}

void SchedProcess::GenCntrLog(LogData* ld, bool gen) {
  if(super_proc) {
    super_proc->GenCntrLog(ld, true);
  }
  if(cntr && gen) {
    if(cntr_items.size < 1)
      GetCntrDataItems();
    ld->AddFloat(cntr_items.FastEl(0), (float)cntr->val);
  }
}

LogData& SchedProcess::GenLogData(LogData* ld) {
  if(ld == NULL)
    ld = &log_data;
  int lc = GetCounter();	// loop counter
  int fc = 0;			// final counter
  if(super_proc != NULL)
    fc = super_proc->GetCounter() + 1; // add 1 to simulate increment of parent log..
  ld->Reset();
  GenCntrLog(ld, log_counter);
  bool no_stats = true;
  if((loop_stats.leaves == 0) && (final_stats.leaves == 0))
    no_stats = false;		// always generate in this case
  Stat* st;
  taLeafItr i;
  FOR_ITR_EL(Stat, st, loop_stats., i) {
    if(log_loop) {
      if(!st->log_stat || !st->mod.flag)
	continue;
      if(((lc - st->mod.off) % (st->mod.m)) == 0) {
	no_stats = false;
	st->GenLogData(ld);
      }
    }
    else {
      no_stats = false;
      st->GenLogData(ld);
    }
  }
  if(!log_loop) {
    FOR_ITR_EL(Stat, st, final_stats., i) {
      if(!st->log_stat || !st->mod.flag)
	continue;
      if(((fc - st->mod.off) % (st->mod.m)) == 0) {
	st->GenLogData(ld);
	no_stats = false;
      }
    }
  }
  if(no_stats)
    ld->Reset();
  return *ld;
}

void SchedProcess::UpdateLogs() {
  if(logs.leaves == 0)
    return;
  GenLogData();			// use our own log_data
  if(log_data.items.size == 0)	// no data this time around
    return;
  PDPLog* lg;
  taLeafItr i;
  FOR_ITR_EL(PDPLog, lg, logs., i) {
    lg->NewData(log_data,this);
  }
  log_data.Reset();		// reset to prevent dangling pointers
}

// returns the default process to create given type of stat in
SchedProcess* SchedProcess::Default_StatProc(TypeDef* stat_td, Process_MGroup* procs) {
  TypeDef* tp = &TA_TrialProcess;
  String tp_nm = stat_td->OptionAfter("COMPUTE_IN_");
  if(!tp_nm.empty())
    tp = taMisc::types.FindName(tp_nm);
  if(tp == NULL)
    tp = &TA_TrialProcess;	// not a bad bet, in general..
  int idx;
  SchedProcess* sp = (SchedProcess*)procs->FindLeafType(tp, idx);
  if(sp == NULL)
    return (SchedProcess*)procs->Leaf(procs->leaves-1); // last one, when in doubt..
  SchedProcess* next_sp = (SchedProcess*)procs->Leaf(idx+1);
  if((next_sp != NULL) && (next_sp->InheritsFrom(tp)))	// if two in a row of same type, return lowest (2nd)
    return next_sp;
  return sp;
}

// returns the default stat group within sched proc to create stat in
Stat_Group* SchedProcess::Default_StatGroup(TypeDef* stat_td, SchedProcess* proc) {
  Stat_Group* stat_gp = &(proc->loop_stats);

  // figure out if it should be in loop or final stats of process
  if(((proc->sub_proc == NULL) || stat_td->HasOption("FINAL_STAT")) && 
     !stat_td->HasOption("LOOP_STAT"))
  {
    stat_gp = &(proc->final_stats);
  }
  return stat_gp;
}

//////////////////////////
// 	CycleProcess	//
//////////////////////////

void CycleProcess::Initialize() {
  sub_proc_type = NULL;
}

//////////////////////////
// 	SettleProcess	//
//////////////////////////

void SettleProcess::Initialize() {
  sub_proc_type = &TA_CycleProcess;
  cycle.SetMax(100);
  log_counter = true;
}

void SettleProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(cycle, this);
  TrialProcess* tp = (TrialProcess*)FindSuperProc(&TA_TrialProcess);
  if(tp != NULL) {
    CyclesToSettle* cs = (CyclesToSettle*)tp->loop_stats.FindType(&TA_CyclesToSettle);
    if(cs != NULL)
      cs->settle = this;
  }
}


//////////////////////////
// 	TrialProcess	//
//////////////////////////

void TrialProcess::Initialize() {
  sub_proc_type = NULL;
  cur_event = NULL;
  epoch_proc = NULL;
  enviro_group = NULL;
}

void TrialProcess::CutLinks() {
  epoch_proc = NULL;
  taBase::DelPointer((TAPtr*)&cur_event);
  taBase::DelPointer((TAPtr*)&enviro_group);
  SchedProcess::CutLinks();
}

void TrialProcess::Copy_(const TrialProcess& cp) {
  taBase::SetPointer((TAPtr*)&cur_event, cp.cur_event);
  taBase::SetPointer((TAPtr*)&enviro_group, cp.enviro_group);
}

void TrialProcess::UpdateAfterEdit() {
  SchedProcess::UpdateAfterEdit();
  epoch_proc = (EpochProcess*)FindSuperProc(&TA_EpochProcess);
  if((epoch_proc != NULL) && epoch_proc->InheritsFrom(TA_SequenceEpoch))
    taBase::SetPointer((TAPtr*)&enviro_group,
			((SequenceEpoch*)epoch_proc)->cur_event_gp);
  else if(environment != NULL)
    taBase::SetPointer((TAPtr*)&enviro_group, &(environment->events));
}

void TrialProcess::Init_impl() {
  SchedProcess::Init_impl();
  if(epoch_proc != NULL)
    taBase::SetPointer((TAPtr*)&cur_event, epoch_proc->cur_event);
  if((epoch_proc != NULL) && epoch_proc->InheritsFrom(TA_SequenceEpoch))
    taBase::SetPointer((TAPtr*)&enviro_group,
			((SequenceEpoch*)epoch_proc)->cur_event_gp);
}

//////////////////////////
// 	EpochProcess	//
//////////////////////////

void EpochProcess::Initialize() {
  sub_proc_type = &TA_TrialProcess;
  trial.SetMax(100);
  order = PERMUTED;
  cur_event = NULL;
  wt_update = ON_LINE;
  batch_n = 10;
  enviro_group = NULL;
}

void EpochProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(event_list, this);
  taBase::Own(trial, this);
}
  
void EpochProcess::CutLinks() {
  taBase::DelPointer((TAPtr*)&cur_event);
  taBase::DelPointer((TAPtr*)&enviro_group);
  SchedProcess::CutLinks();
}

void EpochProcess::Copy_(const EpochProcess& cp) {
  trial = cp.trial;
  taBase::SetPointer((TAPtr*)&cur_event,cp.cur_event);
  order = cp.order;
  wt_update = cp.wt_update;
  batch_n = cp.batch_n;
  event_list = cp.event_list;
  taBase::SetPointer((TAPtr*)&enviro_group,cp.enviro_group);
}

void EpochProcess::UpdateAfterEdit() {
  SchedProcess::UpdateAfterEdit();
  if(environment != NULL)
    taBase::SetPointer((TAPtr*)&enviro_group, &(environment->events));
}

void EpochProcess::Init_impl() {
  SchedProcess::Init_impl();
  if((environment == NULL) || (network == NULL))
    return;

  environment->InitEvents();
  environment->UnSetLayers();

  trial.val = 0;
  GetEventList();
  GetCurEvent();
}

void EpochProcess::GetEventList() {
  event_list.Reset();
  trial.max = environment->EventCount();
  int i;
  for(i=0; i<trial.max; i++)
    event_list.Add(i);
  if(order == PERMUTED)
    event_list.Permute();
}

void EpochProcess::GetCurEvent() {
  if(environment == NULL) return;

  if(trial.max != environment->EventCount()) {
    GetEventList();
    if(trial.val >= trial.max) {
      taBase::SetPointer((TAPtr*)&cur_event, NULL);
      return;
    }
  }

  if((order == PERMUTED) || (order == SEQUENTIAL)) {
    taBase::SetPointer((TAPtr*)&cur_event,
			environment->GetEvent(event_list[trial.val]));
  }
  else {
    long evnt_no = lrand48() % trial.max;
    taBase::SetPointer((TAPtr*)&cur_event, environment->GetEvent((int)evnt_no));
  }
}  

void EpochProcess::Loop() {
  if(cur_event == NULL) return;

  if(trial.max != environment->EventCount()) { // something changed!
    GetEventList();
    if(trial.val >= trial.max) {
      taBase::SetPointer((TAPtr*)&cur_event, NULL);
      return;
    }
    GetCurEvent();		// get the new event
  }
  if(sub_proc != NULL)
    sub_proc->Run();
  if(wt_update == ON_LINE)
    network->UpdateWeights();
  else if((wt_update == SMALL_BATCH) && (((trial.val + 1) % batch_n) == 0))
    network->UpdateWeights();
}

void EpochProcess::UpdateState() {
  SchedProcess::UpdateState();
  GetCurEvent();
}

bool EpochProcess::Crit() {
  bool b = SchedProcess::Crit();
  if(b || (cur_event == NULL))
    return true;
  return false;
}

void EpochProcess::Final() {
  if(wt_update == BATCH)
    network->UpdateWeights();
  else if((wt_update == SMALL_BATCH) && ((trial.val % batch_n) != 0))
    network->UpdateWeights();	// we didn't do this just before end of trial..
}

void EpochProcess::GetCntrDataItems() {
  if(cntr_items.size >= 2)
    return;
  SchedProcess::GetCntrDataItems();
  DataItem* it = (DataItem*)cntr_items.New(1,&TA_DataItem);
  it->SetStringName("Event");
}
  
void EpochProcess::GenCntrLog(LogData* ld, bool gen) {
  SchedProcess::GenCntrLog(ld, gen);
  if(gen) {
    if(cntr_items.size < 2)
      GetCntrDataItems();
    if(cur_event != NULL)
      ld->AddString(cntr_items.FastEl(1), cur_event->name); 
    else
      ld->AddString(cntr_items.FastEl(1), "n/a");
  }
}


//////////////////////////////////
// 	SequenceProcess		//
//////////////////////////////////

void SequenceProcess::Initialize() {
  sub_proc_type = &TA_TrialProcess;
  tick.SetMax(10);
  order = SEQUENTIAL;
  sequence_init = INIT_STATE;
  sequence_epoch = NULL;
  cur_event_gp = NULL;
  cur_event = NULL;
  enviro_group = NULL;
}

void SequenceProcess::Destroy() {
  CutLinks();
}

void SequenceProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(tick, this);
}

void SequenceProcess::CutLinks() {
  sequence_epoch = NULL;
  taBase::DelPointer((TAPtr*)&cur_event_gp);
  taBase::DelPointer((TAPtr*)&cur_event);
  taBase::DelPointer((TAPtr*)&enviro_group);
  SchedProcess::CutLinks();
}

void SequenceProcess::Copy_(const SequenceProcess& cp) {
  tick = cp.tick;
  taBase::SetPointer((TAPtr*)&cur_event,cp.cur_event);
  taBase::SetPointer((TAPtr*)&cur_event_gp,cp.cur_event_gp);
  order = cp.order;
  sequence_init = cp.sequence_init;
  event_list = cp.event_list;
  taBase::SetPointer((TAPtr*)&enviro_group,cp.enviro_group);
}

void SequenceProcess::UpdateAfterEdit() {
  SchedProcess::UpdateAfterEdit();
  sequence_epoch = (SequenceEpoch*)FindSuperProc(&TA_SequenceEpoch);
  if(environment != NULL)
    taBase::SetPointer((TAPtr*)&enviro_group, &(environment->events));
}

void SequenceProcess::Init_impl() {
  SchedProcess::Init_impl();
  if((environment == NULL) || (network == NULL) || (sequence_epoch == NULL))
    return;
  taBase::SetPointer((TAPtr*)&cur_event_gp, sequence_epoch->cur_event_gp);
  if(cur_event_gp == NULL)
    return;

  tick.val = 0;
  GetEventList();
  GetCurEvent();
  InitNetState();
}

void SequenceProcess::GetEventList() {
  event_list.Reset();
  tick.max = cur_event_gp->EventCount();
  int i;
  for(i=0; i<tick.max; i++)
    event_list.Add(i);
  if(order == PERMUTED)
    event_list.Permute();
}

void SequenceProcess::GetCurEvent() {
  if((cur_event_gp == NULL) || (sequence_epoch == NULL))
    return;

  if(tick.max != cur_event_gp->EventCount()) {
    GetEventList();
    if(tick.val >= tick.max) {
      taBase::SetPointer((TAPtr*)&cur_event, NULL);
      return;
    }
  }

  if((order == PERMUTED) || (order == SEQUENTIAL)) {
    taBase::SetPointer((TAPtr*)&cur_event,
			(Event*)cur_event_gp->GetEvent(event_list[tick.val]));
  }
  else {
    long evnt_no = lrand48() % tick.max;
    taBase::SetPointer((TAPtr*)&cur_event,
			(Event*)cur_event_gp->GetEvent((int)evnt_no));
  }
  // copy cur_event to epoch so that trial process can access it there..
  taBase::SetPointer((TAPtr*)&(sequence_epoch->cur_event), cur_event);
}

void SequenceProcess::Loop() {
  if((cur_event_gp == NULL) || (sequence_epoch == NULL))
    return;
  if(tick.max != cur_event_gp->EventCount()) {
    GetEventList();
    if(tick.val >= tick.max) {
      taBase::SetPointer((TAPtr*)&cur_event, NULL);
      return;
    }
    GetCurEvent();
  }
  if(sub_proc != NULL)
    sub_proc->Run();
  if(sequence_epoch->wt_update == EpochProcess::ON_LINE)
    network->UpdateWeights();
}

void SequenceProcess::UpdateState() {
  SchedProcess::UpdateState();
  GetCurEvent();
}

bool SequenceProcess::Crit() {
  bool b = SchedProcess::Crit();
  if(b || (cur_event_gp == NULL))
    return true;
  return false;
}


void SequenceProcess::InitNetState() {
  if(sequence_init == INIT_STATE)
    network->InitState();
  else if(sequence_init == MODIFY_STATE)
    network->ModifyState();
}

void SequenceProcess::GetCntrDataItems() {
  if(cntr_items.size >= 2)
    return;
  SchedProcess::GetCntrDataItems();
  DataItem* it = (DataItem*)cntr_items.New(1,&TA_DataItem);
  it->SetStringName("Event");
}
  
void SequenceProcess::GenCntrLog(LogData* ld, bool gen) {
  SchedProcess::GenCntrLog(ld, gen);
  if(gen) {
    if(cntr_items.size < 2)
      GetCntrDataItems();
    if(cur_event != NULL)
      ld->AddString(cntr_items.FastEl(1), cur_event->name); 
    else
      ld->AddString(cntr_items.FastEl(1), "n/a");
  }
}


//////////////////////////////////
// 	SequenceEpoch		//
//////////////////////////////////

void SequenceEpoch::Initialize() {
  sub_proc_type = &TA_SequenceProcess;
  cur_event_gp = NULL;
}

void SequenceEpoch::Destroy() {
  CutLinks();
}

void SequenceEpoch::CutLinks() {
  taBase::DelPointer((TAPtr*)&cur_event_gp);
  EpochProcess::CutLinks();
}

void SequenceEpoch::Copy_(const SequenceEpoch& cp) {
  taBase::SetPointer((TAPtr*)&cur_event_gp, cp.cur_event_gp);
}

void SequenceEpoch::Init_impl() {
  SchedProcess::Init_impl();
  if((environment == NULL) || (network == NULL))
    return;

  environment->InitEvents();
  environment->UnSetLayers();

  trial.val = 0;
  GetEventList();
  GetCurEvent();
}

void SequenceEpoch::GetEventList() {
  event_list.Reset();
  trial.max = environment->GroupCount();
  if(trial.max == 0) trial.max = 1; // just present all the events themselves!
  int i;
  for(i=0; i<trial.max; i++)
    event_list.Add(i);
  if(order == PERMUTED)
    event_list.Permute();
}

void SequenceEpoch::GetCurEvent() {
  if(environment == NULL) return;

  if((trial.max == 1) && (environment->GroupCount() == 0)) {
    // group = master event list
    taBase::SetPointer((TAPtr*)&cur_event_gp, &(environment->events));
    return;
  }

  if(trial.max != environment->GroupCount()) {
    GetEventList();
    if(trial.val >= trial.max) {
      taBase::SetPointer((TAPtr*)&cur_event_gp, NULL);
      return;
    }
  }

  if((order == PERMUTED) || (order == SEQUENTIAL)) {
    taBase::SetPointer((TAPtr*)&cur_event_gp,
			environment->GetGroup(event_list[trial.val]));
  }
  else {
    long evnt_no = lrand48() % trial.max;
    taBase::SetPointer((TAPtr*)&cur_event_gp, environment->GetGroup((int)evnt_no));
  }
}  
  
void SequenceEpoch::Loop() {
  if(cur_event_gp == NULL) return;

  if((trial.max != environment->GroupCount()) && 
      !((trial.max == 1) && (environment->GroupCount() == 0))) {
    GetEventList();
    if(trial.val >= trial.max) {
      taBase::SetPointer((TAPtr*)&cur_event_gp, NULL);
      return;
    }
    GetCurEvent();
  }
  if(sub_proc != NULL)
    sub_proc->Run();
  if(wt_update == SMALL_BATCH) // small batch = end of sequence
    network->UpdateWeights();
}

bool SequenceEpoch::Crit() {
  bool b = SchedProcess::Crit();
  if(b || (cur_event_gp == NULL))
    return true;
  return false;
}

void SequenceEpoch::Final() {
  if(wt_update == BATCH)
    network->UpdateWeights();
}

void SequenceEpoch::GetCntrDataItems() {
  if(cntr_items.size >= 2)
    return;
  SchedProcess::GetCntrDataItems();
  DataItem* it = (DataItem*)cntr_items.New(1,&TA_DataItem);
  it->SetStringName("EventGp");
}
  
void SequenceEpoch::GenCntrLog(LogData* ld, bool gen) {
  SchedProcess::GenCntrLog(ld, gen);
  if(gen) {
    if(cntr_items.size < 2)
      GetCntrDataItems();
    if((cur_event_gp != NULL) && (!cur_event_gp->name.empty()))
      ld->AddString(cntr_items.FastEl(1), cur_event_gp->name); 
    else
      ld->AddString(cntr_items.FastEl(1), "n/a");
  }
}


//////////////////////////
// 	NEpochProcess	//
//////////////////////////

void NEpochProcess::Initialize() {
  sub_proc_type = &TA_EpochProcess;
  epc_ctr.SetMax(50);
  epoch_proc = NULL;
  log_counter = true;
  epoch = 0;
}

void NEpochProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(epc_ctr, this);
}

void NEpochProcess::CutLinks() {
  epoch_proc = NULL;
  SchedProcess::CutLinks();
}

void NEpochProcess::UpdateAfterEdit() {
  SchedProcess::UpdateAfterEdit();
  epoch_proc = (EpochProcess*)FindSubProc(&TA_EpochProcess);
  if(network != NULL)
    epoch = network->epoch;
}

void NEpochProcess::Init_impl() {
  SchedProcess::Init_impl();
  if(network != NULL)
    epoch = network->epoch;
}  

void NEpochProcess::GetCntrDataItems() {
  if(cntr_items.size >= 2)
    return;
  SchedProcess::GetCntrDataItems();
  DataItem* it = (DataItem*)cntr_items.New(1,&TA_DataItem);
  it->SetNarrowName("Epoch");
}
  
void NEpochProcess::GenCntrLog(LogData* ld, bool gen) {
  SchedProcess::GenCntrLog(ld, gen);
  if(gen) {
    if(cntr_items.size < 2)
      GetCntrDataItems();
    ld->AddFloat(cntr_items.FastEl(1), (float)epoch); 
  }
}

void NEpochProcess::UpdateCounters() {
  SchedProcess::UpdateCounters();
  if(network == NULL)	return;
  if((epoch_proc != NULL) && (epoch_proc->wt_update != EpochProcess::TEST))
    network->epoch++;
  epoch = network->epoch;
}

//////////////////////////
// 	TrainProcess	//
//////////////////////////

void TrainProcess::Initialize() {
  sub_proc_type = &TA_EpochProcess;
  epoch.SetMax(1000);
  epoch_proc = NULL;
  log_counter = true;
}

void TrainProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(epoch, this);
}

void TrainProcess::CutLinks() {
  epoch_proc = NULL;
  SchedProcess::CutLinks();
}

void TrainProcess::UpdateAfterEdit() {
  SchedProcess::UpdateAfterEdit();
  epoch_proc = (EpochProcess*)FindSubProc(&TA_EpochProcess);
  if(network != NULL)
    epoch.val = network->epoch;
}

void TrainProcess::InitAll() {
  Init_force();
//  InitNetwork();   // this is done by the init_force anyway..
  InitAllLogs();
}

void TrainProcess::Init_impl() {
  if(network != NULL) {		// always init network before counter!
    if(network->re_init)
      network->InitWtState();	// this is what it means to init a train process..
  }
  SchedProcess::Init_impl();
}

void TrainProcess::SetReInit(bool ri_val) {
  SchedProcess::SetReInit(ri_val);
  if(network != NULL)
    network->re_init = ri_val;
}

void TrainProcess::UpdateCounters() {
  SchedProcess::UpdateCounters();
  if(network == NULL)	return;
  if((epoch_proc != NULL) && (epoch_proc->wt_update != EpochProcess::TEST))
    network->epoch++;
  epoch.val = network->epoch;
}

void TrainProcess::GetCntrDataItems() {
  SchedProcess::GetCntrDataItems();
}

void TrainProcess::GenCntrLog(LogData* ld, bool gen) {
  SchedProcess::GenCntrLog(ld, gen);
}

void TrainProcess::Run_gui() {
  SchedProcess::Run_gui();
  SetReInit(false); 	// never re-init automatically from gui (only by way of newinit, etc)
}

void TrainProcess::Step_gui() {
  SchedProcess::Step_gui();
  SetReInit(false); 	// never re-init automatically from gui (only by way of newinit, etc)
}


//////////////////////////
// 	BatchProcess	//
//////////////////////////

void BatchProcess::Initialize() {
  sub_proc_type = &TA_TrainProcess;
  batch.SetMax(10);
}

void BatchProcess::InitLinks() {
  SchedProcess::InitLinks();
  taBase::Own(batch, this);
}

