/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef sched_proc_h
#define sched_proc_h

#include <pdp/stats.h>
#include <pdp/enviro.h>
#include <pdp/pdplog.h>


class Counter : public taBase {
  // #INLINE ##NO_TOKENS #NO_UPDATE_AFTER Holds the value of a loop counter
public:
  String	name;			// #HIDDEN not an taNBase to hide name
  int 		val;			// #IV_READ_ONLY #SHOW value of the counter
  int 		max;			// maximum value of the counter

  virtual bool 	Crit()		  
  { if(max >= 0) return (val >= max); else return false; }
  virtual void 	Inc()		  	{ val++; }
  virtual void 	operator=(int i) 	{ val = i; }
  virtual bool  operator==(int i)	{ return val==i; }
  virtual bool  operator<(int i) 	{ return val<i; }
  virtual bool  operator>(int i) 	{ return val>i; }
  virtual bool  operator>=(int i)	{ return val>=i; }  virtual bool  operator<=(int i)	{ return val<=i; }
  virtual bool  operator!=(int i)	{ return val!=i; }
  virtual void 	SetMax(int i)	  	{ max = i; }

  void 	Initialize()		{ val = 0; max = 1; }
  void 	Destroy()		{ };
  void	Copy_(const Counter& cp)	{ name = cp.name; val = cp.val; max = cp.max; }
  COPY_FUNS(Counter, taBase);
  TA_BASEFUNS(Counter);
};


class StepParams : public taBase {
  // #INLINE ##NO_TOKENS #NO_UPDATE_AFTER Holds steping process parameters
public:
  SchedProcess* owner;		// #READ_ONLY #NO_SAVE use this to find the subtypes
  SchedProcess*	proc;		// #SUBTYPE_SchedProcess process to step
  int		n;		// number of steps to take

  TAPtr 	GetOwner() const;
  TAPtr		GetOwner(TypeDef* tp) const;
  TAPtr 	SetOwner(TAPtr ta);

  void	Initialize();
  void	Destroy()	{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const StepParams& cp);
  COPY_FUNS(StepParams, taBase);
  TA_BASEFUNS(StepParams);
};
  

class SchedProcess : public Process {
  // ##MEMB_IN_GPMENU Generic scheduling, looping process
public:
  bool		step_flag;		// #READ_ONLY set by step_proc, not user
  Counter*	cntr;			// #READ_ONLY #NO_SAVE pointer to the counter
  DataItem_List	cntr_items;		// #READ_ONLY data item representations for cntrs
  SchedProcess*	super_proc;		// #READ_ONLY #NO_SAVE #NO_SUBTYPE Process which calls this 
  bool          re_init; 		// #HIDDEN Flag for re-initialization
  bool          can_stop; 		// #HIDDEN true if this process can be stopped
  
  TypeDef*	sub_proc_type;		// #DETAIL #TYPE_SchedProcess #NULL_OK type of sub-process to use
  SchedProcess*	sub_proc;		// #DETAIL Sub-process called by this
  StepParams	step;			// #CONTROL_PANEL Which process to step and n_steps

  Stat_Group	loop_stats;  		// #DETAIL #IN_GPMENU Statistics Computed in Loop
  Stat_Group	final_stats;  		// #DETAIL #IN_GPMENU Statistics Computed after Loop
  Process_Group	init_procs;		// #DETAIL #IN_GPMENU Misc procs run when process is initialized
  Process_Group loop_procs;		// #DETAIL #IN_GPMENU Misc procs run in the loop, using mod based on loop counter
  Process_Group	final_procs;		// #DETAIL #IN_GPMENU Misc procs run after process is finished
  WinView_Group	displays;		// #LINK_GROUP #HIDDEN views to update
  PDPLog_Group	logs;			// #LINK_GROUP #HIDDEN Logs to log to
  bool		log_loop;		// Log the loop state (instead of final state)
  bool		log_counter;		// Log the counter values for this process

  void		NewInit();
  void		ReInit();
  void 		Init();			// initializes process state variables (and sub_procs)
  virtual void	Init_impl();		// #IGNORE this actually has the init code
  virtual void	Init_flag();		// flag this process and all sub procs for re_init
  virtual void	Init_force();		// force re_init of all processes
  virtual void	Init_step_flags();	// initialize step flags
  virtual void	SetReInit(bool ri_val)
  { re_init = ri_val; }			// set re-init flag to given value
  virtual void	InitProcs();		// run the initialize processes (init_procs)

  void		Run_gui();
  void		Step_gui();
  virtual void	NotifyNetObjects();	// #IGNORE notify all network objects of modification
  void 		Step();	
  virtual void	GoTo(int goto_val);	// #BUTTON #GHOST_OFF_running go to given specific counter value

  void 		C_Code(); 		// orchestrates the whole loop

  // inside the loop
  virtual void	Loop();			// called in the do loop (the main looping code)
  virtual void  UpdateCounters();	// update process counters (right after loop)
  virtual void 	LoopProcs();		// call the loop_procs
  virtual void 	LoopStats();		// stats computed inside the loop
  virtual void	UpdateLogs();		// update the logs
  virtual void  UpdateState();		// update misc state variables (after logs)
  virtual void	UpdateDisplays();	// after end of loop
  virtual void	StepCheck();		// check for stepping
  virtual void	StopCheck();		// after all is done, see if stop event pressed
  virtual bool  Crit();			// returns true if stop criterion was met

  // after the loop
  virtual void 	FinalProcs();		// call the final_procs
  virtual void	FinalStats();		// compute the final statistics
  virtual void	Final() {};		// any misc things to do after loop

  virtual int  	GetCounter()		{ if (cntr) return cntr->val; else return 0; }

  LogData& 	GenLogData(LogData* ld=NULL); // generate log data
  virtual void	GetCntrDataItems();	      // #IGNORE get data items for counters
  virtual void 	GenCntrLog(LogData* ld, bool gen); // generate log data for the counters

  virtual SchedProcess* FindSubProc(TypeDef* td); // find of a given type
  virtual SchedProcess* FindSuperProc(TypeDef* td); // find of a given type
  virtual SchedProcess* FindProcOfType(TypeDef* td);
  // find of a given type, including this proc, super and sub procs
  virtual SchedProcess* GetTopProc(); // #IGNORE get the highest-level process

  virtual void	CreateSubProcs(bool update=true);
  // #NEW_FUN #ARGC_0 create sub-processes of this one

  virtual void	InitMyLogs();
  // #MENU #MENU_ON_Actions #MENU_SEP_BEFORE clear all logs associated with this process
  virtual void	InitAllLogs();
  // #MENU #CONFIRM clear all logs that exist in the project
  virtual void	InitNetwork();
  // #MENU #CONFIRM initialize network's weights
  virtual void	InitAll();
  // #MENU #CONFIRM initialize process, network weights, and logs
  virtual void	RemoveFromLogs();
  // #MENU #CONFIRM #MENU_SEP_BEFORE remove this from any logs 
  virtual void	RemoveFromDisplays();
  // #MENU #CONFIRM remove this from any displays
  virtual bool	CheckAllTypes();
  // #MENU #USE_RVAL #MENU_SEP_BEFORE Check all minimum types for all processes and sub_procs
  virtual bool	CheckAllTypes_impl();
  // #IGNORE implementation of check all types 

  // simple process manipulation subroutines to make life a little easier
  virtual void	UnLinkSubProc(); // #IGNORE disconnect from sub process
  virtual void	UnLinkFmSuperProc(); // #IGNORE disconnect me from super process
  virtual void	LinkSubProc(SchedProcess* proc); // #IGNORE link proc object as new sub process (not deleting old one)
  virtual void	LinkSuperProc(SchedProcess* proc); // #IGNORE link proc object as new super process (not deleting old one)
  virtual void	GetAggsFmSuperProc(); // #IGNORE fill in my aggregators from aggs on my super proc
  virtual void	SetAggsAfterRmvSubProc(); // #IGNORE set my aggs to point to my new sub proc stats after removing old one

  // for change my type -- redo all the linking!
  void	ReplacePointersHook(TAPtr old);

  virtual void	MoveToSubGp(const char* gp_name);
  // #MENU #MENU_ON_Structure Move this process and all sub-procs to a new sub group (and name it gp_name)
  virtual void	ChangeNameSuffix(const char* new_name_sufx);
  // #MENU change the suffix (after _ ) for each process name 
  virtual void	AddSuperProc(TypeDef* proc_type);
  // #MENU #TYPE_SchedProcess insert a new super process of given type above this one in hierarchy
  virtual void	AddSubProc(TypeDef* proc_type);
  // #MENU #TYPE_SchedProcess insert a new sub process of given type below this one in hierarchy
  virtual void	RemoveSuperProc();
  // #MENU #CONFIRM remove super process above this one in hierarchy
  virtual void	RemoveSubProc();
  // #MENU #CONFIRM remove sub process below this one in hierarchy

  virtual void	UpdateLogUpdaters();
  // make sure logs have us in their updater lists
  virtual void	UpdateDisplayUpdaters();
  // make sure logs have us in their updater lists

  static SchedProcess* 	Default_StatProc(TypeDef* stat_td, Process_MGroup* procs);
  // returns the default process to create given type of stat in
  static Stat_Group* 	Default_StatGroup(TypeDef* stat_td, SchedProcess* proc);
  // returns the default stat group within sched proc to create stat in

  void 	UpdateAfterEdit();
  void 	Initialize();
  void 	Destroy();
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const SchedProcess& cp);
  COPY_FUNS(SchedProcess, Process);
  TA_BASEFUNS(SchedProcess);
};


//////////////////////////
// 	CycleProcess	//
//////////////////////////

class CycleProcess : public SchedProcess {
  // ##AGGOP_SUM Runs one cycle of activation update
public:
  void 	Initialize();
  void	Destroy()		{ };
  TA_BASEFUNS(CycleProcess);
};


//////////////////////////
// 	SettleProcess	//
//////////////////////////

class SettleProcess : public SchedProcess {
  // ##AGGOP_SUM Settles over cycles of activation propagation
public: 
  Counter	cycle;			// Current cycle number

  void 	Initialize();
  void 	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const SettleProcess& cp)	{ cycle = cp.cycle; }
  COPY_FUNS(SettleProcess, SchedProcess);
  TA_BASEFUNS(SettleProcess);
};


//////////////////////////
// 	TrialProcess	//
//////////////////////////

class TrialProcess : public SchedProcess {
  // ##AGGOP_SUM Runs a single trial (one event)
public:
  Event* 	cur_event;
  // #FROM_GROUP_enviro_group the current event (copied from the EpochProc)
  EpochProcess* epoch_proc;	
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to parent epoch_proc
  Event_MGroup*	enviro_group;	// #READ_ONLY #NO_SAVE main event group on environment

  void 		Init_impl();
  bool		Crit()		{ return true; } // prevent endless looping

  void 	UpdateAfterEdit();	// modify to update the epoch_proc
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	CutLinks();
  void	Copy_(const TrialProcess& cp);
  COPY_FUNS(TrialProcess, SchedProcess);
  TA_BASEFUNS(TrialProcess);
};


//////////////////////////
// 	EpochProcess	//
//////////////////////////

class EpochProcess : public SchedProcess {
  // ##AGGOP_SUM Runs over trials, flat list view of enviro
public:
  enum Order {
    SEQUENTIAL,			// present events in sequential order
    PERMUTED,			// permute the order of event presentation
    RANDOM 			// pick an event at random (with replacement)
  };

  enum WtUpdate {
    TEST,			// don't update weights at all (for testing net)
    ON_LINE,			// update weights on-line (after every event)
    BATCH,			// update weights in batch (after every epoch)
    SMALL_BATCH 		// update weights every batch_n events (or by sequence)
  };

  Counter	trial;		// Current trial number (and index into list)
  Event* 	cur_event;	
  // #FROM_GROUP_enviro_group the current event
  Order		order;		// order to present events in
  WtUpdate	wt_update;	// determines weight update mode
  int		batch_n;	// batch size for small_batch

  int_Array	event_list;	// #HIDDEN list of events
  Event_MGroup*	enviro_group;	// #READ_ONLY #NO_SAVE main event group on environment

  void 		Init_impl();
  void		Loop();		// do on-line, small_batch here
  void		UpdateState();
  bool 		Crit();
  void		Final();	// batch wt update here

  virtual void	GetEventList();	// get list of events from environment
  virtual void	GetCurEvent();	// get current event 

  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  void	UpdateAfterEdit();
  void 	Initialize();	
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const EpochProcess& cp);
  COPY_FUNS(EpochProcess, SchedProcess);
  TA_BASEFUNS(EpochProcess);
};

class SequenceProcess : public SchedProcess {
  // ##AGGOP_SUM Runs a sequence of trials in an event group
public:
  enum Order {
    SEQUENTIAL,			// present events in sequential order
    PERMUTED,			// permute the order of event presentation
    RANDOM 			// pick an event at random (with replacement)
  };

  enum StateInit {		// ways of initializing the state of the network
    DO_NOTHING,			// do nothing
    INIT_STATE,			// initialize the network state
    MODIFY_STATE 		// modify state (algorithm specific)
  };

  Counter	tick;		// one unit of time in a sequence
  Event*	cur_event;	// #FROM_GROUP_cur_event_gp current event
  Event_MGroup*	cur_event_gp;	// #FROM_GROUP_enviro_group event group
  Order		order;		// order to present events in
  StateInit	sequence_init;	// how to initialize network stat at start of sequence

  SequenceEpoch* sequence_epoch;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to parent epoch process
  int_Array	event_list;	// #HIDDEN list of events
  Event_MGroup*	enviro_group;	// #READ_ONLY #NO_SAVE main event group on environment

  void		Init_impl();
  void		Loop();
  void		UpdateState();
  bool 		Crit();

  virtual void	GetEventList();	// get list of events from environment
  virtual void	GetCurEvent();
  virtual void	InitNetState();

  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  void 	UpdateAfterEdit();	// update the sequence_epoch
  void 	Initialize();	
  void 	Destroy();
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const SequenceProcess& cp);
  COPY_FUNS(SequenceProcess, SchedProcess);
  TA_BASEFUNS(SequenceProcess);
};


class SequenceEpoch : public EpochProcess {
  // Runs sequences, uses group view of enviro
public:
  Event_MGroup*	cur_event_gp;	// #FROM_GROUP_enviro_group current event group

  void		Init_impl();
  void		Loop();		// small-batch is update after a sequence
  bool		Crit();
  void		Final();

  void		GetEventList();
  void		GetCurEvent();

  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);
  
  void	Initialize();
  void	Destroy();
  void	CutLinks();
  void	Copy_(const SequenceEpoch& cp);
  COPY_FUNS(SequenceEpoch, EpochProcess);
  TA_BASEFUNS(SequenceEpoch);
};

//////////////////////////
// 	TrainProcess	//
//////////////////////////

class NEpochProcess : public SchedProcess {
  // ##AGGOP_LAST Runs epochs to train network
public:
  Counter	epc_ctr; 	// local epoch counter
  int		epoch;		// epoch count on the network itself
  EpochProcess* epoch_proc;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to epoch process

  void		Init_impl();		// get epoch
  void 		UpdateCounters();	// also update the network epoch
  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  void	UpdateAfterEdit();		// get the epoch_proc
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const NEpochProcess& cp)		{ epoch = cp.epoch; }
  COPY_FUNS(NEpochProcess, SchedProcess);
  TA_BASEFUNS(NEpochProcess);
};

class TrainProcess : public SchedProcess {
  // ##AGGOP_LAST Runs epochs to train network
public:
  Counter	epoch; 			// Epoch Counter
  EpochProcess* epoch_proc;
  // #NO_SUBTYPE #READ_ONLY #NO_SAVE pointer to epoch process

  void		Init_impl();		// also initialize network
  void		SetReInit(bool ri_val);
  void 		UpdateCounters();	// also update the network epoch
  void		GetCntrDataItems();
  void		GenCntrLog(LogData* ld, bool gen);

  void		InitAll();

  void		Run_gui();
  void		Step_gui();

  void	UpdateAfterEdit();		// get the epoch_proc
  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const TrainProcess& cp)		{ epoch = cp.epoch; }
  COPY_FUNS(TrainProcess, SchedProcess);
  TA_BASEFUNS(TrainProcess);
};


//////////////////////////
// 	BatchProcess	//
//////////////////////////

class BatchProcess : public SchedProcess {
  // ##AGGOP_LAST Runs multiple trainings
public:
  Counter	batch;		// number of batches run

  void	Initialize();
  void 	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const BatchProcess& cp)		{ batch = cp.batch; }
  COPY_FUNS(BatchProcess, SchedProcess);
  TA_BASEFUNS(BatchProcess);
};

#endif // sched_proc_h
