/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#include <ta_misc/fun_lookup.h>
#include <math.h>
#include <values.h>

//////////////////////////
//  	FunLookup	//
//////////////////////////

void FunLookup::Initialize() {
  res = .001f;
  x_range.min = 0.0f;
  x_range.max = 1.0f;
  x_range.UpdateAfterEdit();
}

void FunLookup::InitLinks() {
  float_RArray::InitLinks();
  taBase::Own(x_range, this);
}

void FunLookup::Copy_(const FunLookup& cp) {
  x_range = cp.x_range;
  res = cp.res;
}

float FunLookup::Eval(float x) {
  int idx = (int) ((x - x_range.min) / res);
  if(idx < 0) return -MAXFLOAT;
  if(idx >= size) return MAXFLOAT;
  if(idx == size-1) return FastEl(size-1);
  float x_1 = x_range.min + (res * (float)idx);
  float y_1 = FastEl(idx);
  float	y_2 = FastEl(idx+1);
  return y_1 + (y_2 - y_1) * ((x - x_1) / res);
}

void FunLookup::AllocForRange() {
  // range is inclusive -- add some extra..
  x_range.UpdateAfterEdit();
  int sz = (int) (x_range.range / res) + 2;
  Alloc(sz);
  size = sz;
}

void FunLookup::Plot(ostream& strm) {
  strm << "_H:\tx\ty\n";
  int i;
  for(i=0;i<size;i++) {
    float xval = x_range.min + ((float)i * res);
    strm << "_D:\t" << xval << "\t" << El(i) << "\n";
  }
}

void FunLookup::Convolve(const FunLookup& src, const FunLookup& con) {
  x_range = src.x_range;
  res = src.res;
  x_range.min -= con.x_range.min;
  x_range.max -= con.x_range.max;
  x_range.UpdateAfterEdit();
  AllocForRange();
  int ti;
  for(ti=0; ti < size; ti++) {
    float val = 0;
    int ci;
    for(ci=0; ci<con.size; ci++)
      val += src.El(ti + ci) * con.FastEl(ci);
    FastEl(ti) = val;
  }
}
